import logging

from kestrel.exceptions import InvalidECGPattern

_logger = logging.getLogger(__name__)


class Reference:
    def __init__(self, variable, attribute):
        self.variable = variable
        self.attribute = attribute

    def __eq__(self, other):
        if self.variable == other.variable and self.attribute == other.attribute:
            return True
        else:
            return False

    def __str__(self):
        return "{" + self.variable + "." + self.attribute + "}"


def deref_and_flatten_value_to_list(value, deref_func, get_timerange_func):
    # always return a list of values, plus the time range: None or (start, end)
    if isinstance(value, Reference):
        tr = get_timerange_func(value)
        vs = deref_func(value)
        _logger.debug(f"derefed {value} to {vs} with extracted timerange {tr}")
        return vs, tr
    elif isinstance(value, list):
        xs = []
        start, end = None, None
        for x in value:
            vs, tr = deref_and_flatten_value_to_list(x, deref_func, get_timerange_func)
            xs.extend(vs)
            if tr:
                s, e = tr
                if start is None or s < start:
                    start = s
                if end is None or e > end:
                    end = e
        if start is None and end is None:
            timerange = None
        else:
            timerange = start, end
        return xs, timerange
    else:
        return [value], None


def value_to_stix(value):
    if isinstance(value, str):
        return "'" + value + "'"
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, (list, tuple)):
        return "(" + ",".join(map(value_to_stix, value)) + ")"
    elif isinstance(value, Reference):
        raise KestrelInternalError("reference should be derefed before value_to_stix()")
    else:
        raise InvalidECGPattern(f"invalid term: {value}")



"""The abstract interface for building an analytics interface for Kestrel.

A Kestrel analytics interface is a Python package with the following rules:

- The package name should use prefix ``kestrel_analytics_``.

- The package should have one and only one root level class inherited from
  :class:`AbstractAnalyticsInterface`.

  - There is no restriction on package structure for the package.

  - There is no restriction on interface class name.

  - The interface class should inhert :class:`AbstractAnalyticsInterface`.

  - The interface class should be importable from the package directly, i.e.,
    it needs to be imported into ``__init__.py`` of the package.

  - Zero class inherited from :class:`AbstractAnalyticsInterface` will
    result in an exception.

  - Multiple classes inherited from :class:`AbstractAnalyticsInterface` will
    result in an exception.

"""

from abc import ABC, abstractmethod

MODULE_PREFIX = "kestrel_analytics_"


class AbstractAnalyticsInterface(ABC):
    """The abstract class for building an analytics interface."""

    @staticmethod
    @abstractmethod
    def schemes():
        """``scheme`` (the URI prefix before ``://``) of the analytics interface.

        Every analytics interface should have at least one *unique* scheme to
        use at the beginning of the analytics URI. To develop a new analytics
        interface, one needs to check public Kestrel analytics packages to name a
        new one that is not taken. Note that scheme defined here should be in
        lowercase, and Kestrel analytics manager will normalize schemes of incoming
        URIs into lowercase.

        Returns:
            [str]: A list of schemes; A URI with one of the scheme will be
            processed by this interface.

        """
        return []

    @staticmethod
    @abstractmethod
    def list_analytics(config):
        """List analytics names accessible from this interface.

        Args:
            config (dict): a layered list/dict that contains config for the
              interface and can be edited/updated by the interface.

        Returns:
            [str]: A list of analytics names accessible from this interface.

        """
        return []

    @staticmethod
    @abstractmethod
    def execute(uri, argument_variables, config, session_id=None, parameters=None):
        """Execute an analytics.

        An analytics updates argument variables in place with revised
        attributes or additional attributes computed. Therefore, there is no
        need to return any variable, but the optional display object can be
        returned if the analytics generate anything to be shown by the
        front-end, e.g., a visualization analytics.

        When realizing the execute() method of an analytics interface, one
        needs to realize the following functionalities:

            #. Execute the specified analytics.
            #. Keep track of input/output Kestrel variables
            #. Kestrel variable update (in most cases, this is done by the
               analytics interface; in some cases, an analytics may directly
               update the store).
            #. Prepare returned display object.

        Args:
            uri (str): the full URI including the scheme and analytics name.

            argument_variables ([kestrel.symboltable.variable.VarStruct]): the list of Kestrel variables as arguments.

            config (dict): a layered list/dict that contains config for the
              interface and can be edited/updated by the interface.

            session_id (str): id of the session, may be useful for analytics directly writing into the store.

            parameters (dict): analytics execution parameters in key-value pairs ``{"str":"str"}``.

        Returns:
            kestrel.codegen.display.AbstractDisplay: returned display or None.

        """
        return None



# Python 3.10 supports this by default
# https://stackoverflow.com/questions/36286894/name-not-defined-in-type-annotation
from __future__ import annotations

from typeguard import typechecked

from typing import Tuple, Optional
import datetime
from abc import ABC, abstractmethod
from firepit.query import Filter, Predicate
from firepit.timestamp import timefmt
from kestrel.syntax.utils import merge_timeranges
from kestrel.syntax.reference import (
    Reference,
    deref_and_flatten_value_to_list,
    value_to_stix,
)
from kestrel.exceptions import (
    KestrelNotImplemented,
    InvalidECGPattern,
    KestrelInternalError,
)


class ExtCenteredGraphConstruct(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def to_stix(self, center_entity_type: str):
        pass

    @abstractmethod
    def to_firepit(self, center_entity_type: str):
        pass

    @abstractmethod
    def deref(self, deref_func):
        pass


@typechecked
class ExtCenteredGraphPattern(ExtCenteredGraphConstruct):
    def __init__(self, graph: ExtCenteredGraphConstruct):
        self.graph = graph
        self.timerange = None
        self.center_entity_type = None

    def __str__(self):
        return (
            f"graph: {self.graph}, "
            f"timerange: {self.timerange}, "
            f"center_entity_type: {self.center_entity_type}"
        )

    def add_center_entity(self, center_entity_type: str):
        self.center_entity_type = center_entity_type

    def prune_away_centered_graph(self, center_entity_type):
        # only leave the disconnected extended graph components
        if self.graph is not None:
            if_preserve_func = _make_extract_func(center_entity_type, "ext")
            self.graph = self.graph.prune(if_preserve_func)

    def prune_away_extended_graph(self, center_entity_type):
        # only leave the connected centered graph components
        if self.graph is not None:
            if_preserve_func = _make_extract_func(center_entity_type, "center")
            self.graph = self.graph.prune(if_preserve_func)

    def to_stix(
        self,
        timerange: Optional[Tuple[datetime.datetime, datetime.datetime]],
        timeadj: Optional[Tuple[datetime.timedelta, datetime.timedelta]],
    ):
        # timerange: user-specified timerange from Kestrel command
        # timeadj: time adjustments for START and STOP

        # before calling this, make sure:
        # 1. called deref()
        # 2. called add_center_entity()

        if self.center_entity_type is None:
            raise KestrelInternalError(
                "should run add_center_entity() before to_stix()"
            )

        if self.graph is None:
            inner = ""
        else:
            inner = self.graph.to_stix(self.center_entity_type)
        body = "[" + inner + "]"

        if timerange:
            tr = timerange
        elif self.timerange:
            tr = self.timerange
            if timeadj:
                tr = (tr[0] + timeadj[0], tr[1] + timeadj[1])
        else:
            tr = None

        if tr:
            tr_stix = f" START t'{timefmt(tr[0])}' STOP t'{timefmt(tr[1])}'"
        else:
            tr_stix = ""

        return body + tr_stix

    def to_firepit(self):
        if self.center_entity_type is None:
            raise KestrelInternalError(
                "should run add_center_entity() before to_firepit()"
            )
        if self.graph is None:
            return None
        else:
            return Filter([self.graph.to_firepit(self.center_entity_type)])

    def deref(self, deref_func, get_timerange_func):
        if self.graph is not None:
            self.timerange = self.graph.deref(deref_func, get_timerange_func)

    def extend(self, junction_type: str, other_ecgp: Optional[ExtCenteredGraphPattern]):

        if other_ecgp is not None and other_ecgp.graph is not None:

            if self.center_entity_type is None:
                self.center_entity_type = other_ecgp.center_entity_type
            elif other_ecgp.center_entity_type is None:
                pass
            elif self.center_entity_type != other_ecgp.center_entity_type:
                raise InvalidECGPattern(
                    "could not merge ECGPs with different center entities:"
                    + self.center_entity_type
                    + ", "
                    + other_ecgp.center_entity_type
                )

            self.timerange = merge_timeranges((self.timerange, other_ecgp.timerange))

            junction_type = junction_type.upper()

            if junction_type == "AND":
                self.graph = ECGPJunction("AND", self.graph, other_ecgp.graph)
            elif junction_type == "OR":
                self.graph = ECGPJunction("OR", self.graph, other_ecgp.graph)
            else:
                raise KestrelInternalError(
                    f'Junction type {junction_type} not supported besides "AND", "OR".'
                )


@typechecked
class ECGPJunction(ExtCenteredGraphConstruct):
    def __init__(
        self,
        relation: str,
        lhs: ExtCenteredGraphConstruct,
        rhs: ExtCenteredGraphConstruct,
    ):
        # relation: "AND" | "OR"
        self.lhs = lhs
        self.rhs = rhs
        relation = relation.upper()
        if relation == "AND":
            self.relation = "AND"
        elif relation == "OR":
            self.relation = "OR"
        else:
            raise KestrelInternalError("unsupported relation for ECGPJunction()")

    def __str__(self):
        return f"({self.lhs}) {self.relation} ({self.rhs})"

    def prune(self, if_preserve_func):
        self.lhs = self.lhs.prune(if_preserve_func)
        self.rhs = self.rhs.prune(if_preserve_func)
        if self.lhs is None:
            return self.rhs
        elif self.rhs is None:
            return self.lhs
        else:
            return self

    def to_stix(self, center_entity_type: str):
        return (
            "("
            + self.lhs.to_stix(center_entity_type)
            + " "
            + self.relation
            + " "
            + self.rhs.to_stix(center_entity_type)
            + ")"
        )

    def to_firepit(self, center_entity_type: str):
        return Predicate(
            self.lhs.to_firepit(center_entity_type),
            self.relation,
            self.rhs.to_firepit(center_entity_type),
        )

    def deref(self, deref_func, get_timerange_func):
        ltr = self.lhs.deref(deref_func, get_timerange_func)
        rtr = self.rhs.deref(deref_func, get_timerange_func)
        return merge_timeranges((ltr, rtr))


@typechecked
class ECGPComparison(ExtCenteredGraphConstruct):
    def __init__(self, entity_attribute: str, operator: str, value, entity_type=None):
        self.etype = entity_type
        self.attribute = entity_attribute
        self.op = operator.upper()
        self.value = value
        if isinstance(self.value, list):
            if self.op not in ("IN", "NOT IN"):
                raise InvalidECGPattern(
                    'a list should be paired with the operator "IN"'
                )
        elif not isinstance(self.value, Reference):
            if self.op in ("IN", "NOT IN"):
                raise InvalidECGPattern(
                    'inappropriately pair operator "IN" with literal'
                )

    def __str__(self):
        return f"{self.etype}:{self.attribute} {self.op} {self.value}"

    def prune(self, if_preserve_func):
        if if_preserve_func(self.etype):
            return self
        else:
            return None

    def to_stix(self, center_entity_type: str):
        if not self.etype:
            self.etype = center_entity_type
        return f"{self.etype}:{self.attribute} {self.op} {value_to_stix(self.value)}"

    def to_firepit(self, center_entity_type: str):
        if self.etype and self.etype != center_entity_type:
            raise KestrelNotImplemented("firepit filtering on linked entities")
        return Predicate(self.attribute, self.op, self.value)

    def deref(self, deref_func, get_timerange_func):
        xs, tr = deref_and_flatten_value_to_list(
            self.value, deref_func, get_timerange_func
        )
        if len(xs) == 0:
            raise InvalidECGPattern("empty value after deref of {self.value}")
        elif len(xs) == 1:
            self.value = xs[0]
            if self.op == "IN":
                self.op = "="
            elif self.op == "NOT IN":
                self.op = "!="
        else:
            self.value = xs
            if self.op in ("=", "==", "IN"):
                self.op = "IN"
            elif self.op in ("!=", "NOT IN"):
                self.op = "NOT IN"
            else:
                raise InvalidECGPattern(
                    "operator {self.op} incompatible with value {self.value}"
                )
        return tr


def _make_extract_func(center_entity_type, preserve_center_or_ext: str):
    def if_preserve_func(entity_type):
        if entity_type is None:
            connected_to_center = True
        elif entity_type == center_entity_type:
            connected_to_center = True
        else:
            connected_to_center = False

        if preserve_center_or_ext == "center":
            return connected_to_center
        elif preserve_center_or_ext == "ext":
            return not connected_to_center
        else:
            raise KestrelInternalError(
                "unsupported argument in _make_extract_func: {prune_center_or_ext}"
            )

    return if_preserve_func



"""Python analytics interface executes Python function as Kestrel analytics.

Use a Python Analytics
----------------------

Create a profile for each analytics in the python analytics interface config
file (YAML):

- Default path: ``~/.config/kestrel/pythonanalytics.yaml``.
- A customized path specified in the environment variable ``KESTREL_PYTHON_ANALYTICS_CONFIG``.

Example of the python analytics interface config file:

.. code-block:: yaml

    profiles:
        analytics-name-1: # the analytics name to use in the APPLY command
            module: /home/user/kestrel-analytics/analytics/piniponmap/analytics.py
            func: analytics # the analytics function in the module to call
        analytics-name-2:
            module: /home/user/kestrel-analytics/analytics/suspiciousscoring/analytics.py
            func: analytics

Develop a Python Analytics
--------------------------

A Python analytics is a python function that follows the rules:

#. The function takes in one or more Kestrel variable dumps in Pandas DataFrames.

#. The return of the function is a tuple containing either or both:

    - Updated variables. The number of variables can be either 0, e.g.,
      visualization analytics, or the same number as input Kestrel variables.
      The order of the updated variables should follow the same order as input
      variables.

    - An object to display, which can be any of the following types:

        - Kestrel display object

        - HTML element as a string

        - Matplotlib figure (by default, Pandas DataFrame plots use this)

   The display object can be either before or after updated variables. In other
   words, if the input variables are ``var1``, ``var2``, and ``var3``, the
   return of the analytics can be either of the following:

   .. code-block:: python

       # the analytics enriches variables without returning a display object
       return var1_updated, var3_updated, var3_updated

       # this is a visualization analytics and no variable updates
       return display_obj

       # the analytics does both variable updates and visualization
       return var1_updated, var3_updated, var3_updated, display_obj

       # the analytics does both variable updates and visualization
       return display_obj, var1_updated, var3_updated, var3_updated


#. Parameters in the APPLY command are passed in as environment varibles. The
   names of the environment variables are the exact parameter keys given in the
   ``APPLY`` command. For example, the following command

   .. code-block::

       APPLY python://a1 ON var1 WITH XPARAM=src_ref.value, YPARAM=number_observed

   creates environment variables ``$XPARAM`` with value ``src_ref.value`` and
   ``$YPARAM`` with value ``number_observed`` to be used by the analytics
   ``a1``. After the execution of the analytics, the environment variables will
   be roll back to the original state.

#. The Python function could spawn other processes or execute other binaries,
   where the Python function just acts like a wrapper. Check our `domain name
   lookup analytics`_ as an example.

.. _domain name lookup analytics: https://github.com/opencybersecurityalliance/kestrel-analytics/tree/release/analytics/domainnamelookup

"""

import os
import sys
import pathlib
import logging
import inspect
import traceback
from collections.abc import Mapping
from pandas import DataFrame
from importlib.util import spec_from_file_location, module_from_spec
from contextlib import AbstractContextManager

from kestrel.codegen.display import AbstractDisplay, DisplayHtml, DisplayFigure
from kestrel.analytics import AbstractAnalyticsInterface
from kestrel.exceptions import (
    InvalidAnalytics,
    AnalyticsManagerInternalError,
    InvalidAnalyticsInterfaceImplementation,
    AnalyticsError,
    InvalidAnalyticsArgumentCount,
    InvalidAnalyticsOutput,
)
from kestrel_analytics_python.config import (
    get_profile,
    load_profiles,
)

_logger = logging.getLogger(__name__)


class PythonInterface(AbstractAnalyticsInterface):
    @staticmethod
    def schemes():
        """Python analytics interface only supports ``python://`` scheme."""
        return ["python"]

    @staticmethod
    def list_analytics(config):
        """Load config to list avaliable analytics."""
        if not config:
            config["profiles"] = load_profiles()
        analytics = list(config["profiles"].keys())
        analytics.sort()
        return analytics

    @staticmethod
    def execute(uri, argument_variables, config, session_id=None, parameters=None):
        """Execute an analytics."""
        scheme, _, profile = uri.rpartition("://")

        if not config:
            config["profiles"] = load_profiles()

        if scheme != "python":
            raise AnalyticsManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        with PythonAnalytics(profile, config["profiles"], parameters) as func:
            display = func(argument_variables)

        return display


class PythonAnalytics(AbstractContextManager):
    """Handler of a Python Analytics

    Use it as a context manager:

    .. code-block:: python

        with PythonAnalytics(profile_name, profiles, parameters) as func:
            func(input_kestrel_variables)

    #. Validate and retrieve profile data. The data should be a dict with "module"
       and "func", plus appropriate values.

    #. Prepare the analytics by loading the module. Also verify the function
       exists.

    #. Execute the analytics and process return intelligently.

    #. Clean the environment.

    Args:

        profile_name (str): The name of the profile/analytics.

        profiles (dict): name to profile (dict) mapping.

        parameters (dict): key-value pairs of parameters.

    """

    def __init__(self, profile_name, profiles, parameters):
        self.name = profile_name
        self.parameters = parameters
        self.module_name, self.func_name = get_profile(profile_name, profiles)
        self.module_path = pathlib.Path(self.module_name).expanduser().resolve()
        self.module_path_dir_str = str(self.module_path.parent)

    def __enter__(self):
        # accommodate any other Python modules to load in the dir
        self.syspath = sys.path.copy()
        sys.path.append(self.module_path_dir_str)

        # accommodate any other executables or data to load in the dir
        self.cwd_original = os.getcwd()
        os.chdir(self.module_path_dir_str)

        # passing parameters as environment variables
        self.environ_original = os.environ.copy()
        if self.parameters:
            if isinstance(self.parameters, Mapping):
                _logger.debug(f"setting parameters as env vars: {self.parameters}")
                os.environ.update(self.parameters)
            else:
                raise InvalidAnalyticsInterfaceImplementation(
                    "parameters should be passed in as a Mapping"
                )

        # time to load the analytics function
        self.analytics_function = self._locate_analytics_func(self._load_module())

        return self._execute

    def __exit__(self, exception_type, exception_value, traceback):
        sys.path = self.syspath
        os.chdir(self.cwd_original)
        os.environ = self.environ_original

    def _execute(self, arg_variables):
        """Execute the analytics

        Args:
            arg_variables ([VarStruct]): input variables to the analytics.

        Returns:

            Kestrel Display object: The Display object is the explicit return.
            The function also has side effect to update the session with
            returned Kestrel variables.

        """
        input_dataframes = [DataFrame(v.get_entities()) for v in arg_variables]
        if len(input_dataframes) != self._get_var_count():
            raise InvalidAnalyticsArgumentCount(
                self.name, len(input_dataframes), self._get_var_count()
            )
        else:
            try:
                outputs = self.analytics_function(*input_dataframes)
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                error = "".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )
                raise AnalyticsError(f"{self.name} failed at execution:\n{error}")

            if not isinstance(outputs, tuple):
                outputs = (outputs,)

            output_dfs, output_dsps = [], []
            for x in outputs:
                x_class_str = type(x).__module__ + "." + type(x).__name__
                if isinstance(x, DataFrame):
                    output_dfs.append(x)
                elif isinstance(x, AbstractDisplay):
                    _logger.debug(f'analytics "{self.name}" yielded a display object')
                    output_dsps.append(x)
                elif isinstance(x, str):
                    _logger.info(
                        f'analytics "{self.name}" yielded a string return. treat it as an HTML element.'
                    )
                    output_dsps.append(DisplayHtml(x))
                elif x_class_str == "matplotlib.figure.Figure":
                    _logger.info(f'analytics "{self.name}" yielded a figure.')
                    output_dsps.append(DisplayFigure(x))
                else:
                    raise InvalidAnalyticsOutput(self.name, type(x))

            if not outputs:
                raise AnalyticsError(f'analytics "{self.name}" yield nothing')
            if len(output_dsps) > 1:
                raise AnalyticsError(
                    f'analytics "{self.name}" yielded more than one Kestrel Display object'
                )
            if output_dfs:
                if len(output_dfs) != len(input_dataframes):
                    raise AnalyticsError(
                        f'analytics "{self.name}" yielded less/more Kestrel variable(s) than given'
                    )

                for v, odf in zip(arg_variables, output_dfs):
                    if not "id" in odf:
                        raise AnalyticsError(
                            f'analytics "{self.name}" yielded invalid return without "id"'
                        )
                    if not "type" in odf:
                        raise AnalyticsError(
                            f'analytics "{self.name}" yielded invalid return without "type"'
                        )

                    o_type_set = set(odf["type"])
                    o_type = o_type_set.pop()
                    if o_type_set:
                        raise AnalyticsError(
                            f'analytics "{self.name}" yielded invalid return with inconsistent types'
                        )

                    o_dict = odf.to_dict(orient="records")
                    v.store.reassign(v.entity_table, o_dict)

            display = output_dsps[0] if output_dsps else None
            return display

    def _load_module(self):
        spec = spec_from_file_location(
            "kestrel_analytics_python.analytics.{profile_name}", str(self.module_path)
        )

        try:
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
        except ModuleNotFoundError as e:
            raise AnalyticsError(
                f"{self.name} misses dependent library: {e.name}",
                "pip install the corresponding Python package",
            )
        except Exception as e:
            if isinstance(e, AttributeError) and e.args == (
                "'NoneType' object has no attribute 'loader'",
            ):
                raise AnalyticsError(
                    f"{self.name} is not found",
                    "please make sure the Python module and function specified in the profile (configuration) exist",
                )
            else:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                error = "".join(
                    traceback.format_exception(exc_type, exc_value, exc_traceback)
                )
                raise AnalyticsError(f"{self.name} failed at importing:\n{error}")

        return module

    def _locate_analytics_func(self, module):
        if hasattr(module, self.func_name):
            return getattr(module, self.func_name)
        else:
            raise InvalidAnalytics(
                self.name,
                "python",
                f'function "{self.func_name}" not exist in module: {self.module_path}',
            )

    def _get_var_count(self):
        sig = inspect.signature(self.analytics_function)
        return len(sig.parameters)



"""The STIX bundle data source package provides access to canned data in STIX
bundles locally or remotely.

"""

import json
import logging
import os
import re
import uuid
import pathlib
import shutil
from datetime import datetime, timedelta, timezone
from dateutil import parser
import requests

from stix2matcher.matcher import Pattern

from firepit.woodchipper import convert_to_stix

from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import DataSourceManagerInternalError, DataSourceConnectionError

_logger = logging.getLogger(__name__)


def _make_query_id(uri, pattern):
    return str(uuid.uuid5(uuid.NAMESPACE_URL, str(uri) + pattern))


def _make_query_dir(query_id):
    path = pathlib.Path(query_id)
    path.mkdir(parents=True, exist_ok=False)
    return path


def _make_download_dir():
    path = pathlib.Path("downloads")
    path.mkdir(parents=True, exist_ok=True)
    return path


def _clean_ingestdir_and_raise_error(ingestdir, uri):
    # it is important to clean the directory before raise error
    # otherwise, the next execution will find the dir and assume good data there
    shutil.rmtree(ingestdir)
    raise DataSourceConnectionError(uri)


def fixup_pattern(pattern):
    # The matcher doesn't accept TimestampLiterals in START/STOP
    # See https://github.com/oasis-open/cti-pattern-validator/issues/52
    return re.sub(r"(START|STOP)\s+t'", r"\1 '", pattern)


class StixBundleInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX bundle data source interface supporting ``file:///``, ``http://``, ``https://`` scheme."""
        return ["file", "http", "https"]

    @staticmethod
    def list_data_sources(config=None):
        """This interface does not list data sources."""
        return []

    @staticmethod
    def query(uri, pattern, session_id=None, config=None):
        """Query a STIX bundle locally or remotely."""

        _logger.debug(f"query URI received at interface_stixbundle: {uri}")

        scheme, _, data_paths = uri.rpartition("://")
        data_paths = data_paths.split(",")
        pattern = fixup_pattern(pattern)
        compiled_pattern = Pattern(pattern)
        query_id = _make_query_id(uri, pattern)
        downloaddir = _make_download_dir()

        try:
            ingestdir = _make_query_dir(query_id)
        except FileExistsError:
            # We already cached this bundle
            data_paths = []

        bundles = []
        for i, data_path in enumerate(data_paths):
            _logger.debug(f"requesting data from path: {data_path}")

            if scheme == "file":
                rawfile = data_path
                try:
                    with open(data_path, "r") as f:
                        bundle_in = json.load(f)
                except Exception:
                    _clean_ingestdir_and_raise_error(ingestdir, uri)

            elif scheme == "http" or scheme == "https":
                data_uri = f"{scheme}://{data_path}"
                data_path, extension = os.path.splitext(data_path)
                data_path_stripped = "".join(filter(str.isalnum, data_path))
                rawfile = downloaddir / f"{data_path_stripped}"
                if extension:
                    rawfile = rawfile.with_suffix(f"{extension}")
                last_modified = None
                file_time = None
                if rawfile.exists():
                    _logger.debug("File exists: %s", rawfile)
                    try:
                        resp = requests.head(data_uri)
                    except requests.exceptions.ConnectionError:
                        _clean_ingestdir_and_raise_error(ingestdir, uri)

                    file_time = datetime.fromtimestamp(
                        rawfile.stat().st_mtime, tz=timezone.utc
                    )
                    last_modified = resp.headers.get("Last-Modified")
                    if last_modified:
                        last_modified = parser.parse(last_modified)
                    else:
                        _logger.debug(
                            "HTTP/HTTPS response header does not have 'Last-Modified' field"
                        )
                        last_modified = datetime.now(timezone.utc) - timedelta(
                            minutes=5
                        )
                else:
                    _logger.debug("File not on disk: %s", rawfile)

                if not last_modified or last_modified > file_time:
                    _logger.info("Downloading %s to %s", data_uri, rawfile)
                    try:
                        resp = requests.get(data_uri, stream=True)
                    except requests.exceptions.ConnectionError:
                        _clean_ingestdir_and_raise_error(ingestdir, uri)

                    with rawfile.open("wb") as f:
                        for chunk in resp.iter_content(chunk_size=8192):
                            f.write(chunk)
                else:
                    # We already have this file
                    _logger.debug("Using cached file: %s", rawfile)

                try:
                    bundle_in = _get_bundle(rawfile)
                except Exception:
                    _clean_ingestdir_and_raise_error(ingestdir, uri)
            else:
                raise DataSourceManagerInternalError(
                    f"interface {__package__} should not process scheme {scheme}"
                )

            bundle_out = {}
            _logger.debug("Filtering: %s", rawfile)
            count = 0
            matched = 0
            for prop, val in bundle_in.items():
                if prop == "objects":
                    bundle_out[prop] = []
                    for obj in val:
                        count += 1
                        if obj["type"] != "observed-data" or compiled_pattern.match(
                            [obj], False
                        ):
                            bundle_out[prop].append(obj)
                            matched += 1
                else:
                    bundle_out[prop] = val
            _logger.debug("Matched %d of %d observations: %s", matched, count, rawfile)

            ingestfile = ingestdir / f"{i}.json"
            with ingestfile.open("w") as f:
                json.dump(bundle_out, f)
            bundles.append(str(ingestfile.expanduser().resolve()))

        return ReturnFromFile(query_id, bundles)


def _get_bundle(rawfile):
    try:
        with open(rawfile, "r") as fp:
            bundle = json.load(fp)
    except (json.decoder.JSONDecodeError, UnicodeDecodeError):
        # It's not JSON.  Maybe firepit can convert it to STIX?
        bundle = convert_to_stix(str(rawfile))
    return bundle



import pathlib
import os
import uuid
import collections.abc


def unescape_quoted_string(s):
    return s[1:-1].encode("utf-8").decode("unicode_escape")


def lowered_str_list(xs):
    return [x.lower() for x in xs if isinstance(x, str)]


def update_nested_dict(dict_old, dict_new):
    for k, v in dict_new.items():
        if isinstance(v, collections.abc.Mapping) and k in dict_old:
            dict_old[k] = update_nested_dict(dict_old[k], v)
        else:
            dict_old[k] = v
    return dict_old


def remove_empty_dicts(ds):
    # remove dict with all values as None in list({string:string})
    # this is the results from SQL query
    return [d for d in ds if set(d.values()) != {None}]


def dedup_dicts(ds):
    # deduplicate list({string:string})
    # this is the results from SQL query
    return [dict(s) for s in set(frozenset(d.items()) for d in ds)]


def dedup_ordered_dicts(ds):
    # deduplicate list({string:string})
    # maintain the order if seen
    res = []
    seen = set()
    for d in ds:
        s = str(d)
        if s not in seen:
            res.append(d)
        seen.add(s)
    return res


def subgroup_list(xs, gsize):
    return [xs[i : i + gsize] for i in range(0, len(xs), gsize)]


def mkdtemp():
    # create a temporary directory and return it (named after a random uuid)
    ps = None
    while not ps or pathlib.Path(ps).exists():
        ps = str(uuid.uuid4())
    p = pathlib.Path(ps)
    p.mkdir(parents=True, exist_ok=True)
    return p


class set_current_working_directory:
    def __init__(self, new_cwd):
        self.tmp_cwd = new_cwd

    def __enter__(self):
        self.cwd = os.getcwd()
        os.chdir(self.tmp_cwd)

    def __exit__(self, exception_type, exception_value, traceback):
        os.chdir(self.cwd)



from typeguard import typechecked
from lark import Lark
from pkgutil import get_data
from itertools import chain
from typing import Tuple, Iterable
import datetime

from kestrel.codegen.relations import (
    all_relations,
    stix_2_0_ref_mapping,
    stix_2_0_identical_mapping,
)

LITERALS = {"CNAME", "LETTER", "DIGIT", "WS", "INT", "WORD", "ESCAPED_STRING", "NUMBER"}
AGG_FUNCS = {"MIN", "MAX", "AVG", "SUM", "COUNT", "NUNIQUE"}
TRANSFORMS = {"TIMESTAMPED"}


def get_keywords():
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    parser = Lark(grammar, parser="lalr")
    alphabet_patterns = filter(lambda x: x.pattern.value.isalnum(), parser.terminals)
    keywords = [x.pattern.value for x in alphabet_patterns] + all_relations
    keywords_lower = map(lambda x: x.lower(), keywords)
    keywords_upper = map(lambda x: x.upper(), keywords)
    keywords_comprehensive = list(chain(keywords_lower, keywords_upper))
    return keywords_comprehensive


def get_entity_types():
    all_types = {"x-ibm-finding", "x-oca-asset", "x-oca-event"}
    for mapping in stix_2_0_ref_mapping:
        for i in (0, 2):
            if not mapping[i].endswith("-ext"):
                all_types.add(mapping[i])
    all_types.update(stix_2_0_identical_mapping.keys())
    return tuple(all_types)


def get_all_input_var_names(stmt):
    input_refs = ["input", "input_2", "variablesource"]
    inputs_refs = stmt["inputs"] if "inputs" in stmt else []
    return [stmt.get(k) for k in input_refs if k in stmt] + inputs_refs


@typechecked
def merge_timeranges(trs: Iterable[Tuple[datetime.datetime, datetime.datetime]]):
    # return the earliest start time and latest end time
    trs = [tr for tr in trs if tr is not None]
    if trs:
        starts, ends = list(zip(*trs))
        start = min(starts)
        end = max(ends)
        return (start, end)
    else:
        return None


@typechecked
def timedelta_seconds(t: int):
    return datetime.timedelta(seconds=t)



from kestrel.datasource.interface import MODULE_PREFIX, AbstractDataSourceInterface
from kestrel.datasource.retstruct import ReturnFromFile, ReturnFromStore
from kestrel.datasource.manager import DataSourceManager



"""Relationship mapping from Kestrel relation to STIX references.

STIX reference names may not be the original STIX reference name. The names
used here are pre-processed by :func:`firepit.raft.invert`. Check the function
for more details.

"""

import dateutil.parser
import datetime
from collections import defaultdict
import logging

from kestrel.syntax.reference import value_to_stix
from firepit.query import Column, Join, Query, Projection, Table, Unique

_logger = logging.getLogger(__name__)

stix_2_0_ref_mapping = {
    # (EntityX, Relate, EntityY): ([EntityX_STIX_Ref_i], [EntityY_STIX_Ref_i])
    # All STIX 2.0 refs enumerated
    # file
    ("file", "contained", "artifact"): (["content_ref"], []),
    ("directory", "contained", "directory"): (["contains_refs"], ["contains_refs"]),
    ("directory", "contained", "file"): (["contains_refs"], ["parent_directory_ref"]),
    ("archive-ext", "contained", "file"): (["contains_refs"], []),
    # email
    ("user-account", "owned", "email-addr"): ([], ["belongs_to_ref"]),
    ("email-addr", "created", "email-message"): ([], ["from_ref", "sender_ref"]),
    ("email-addr", "accepted", "email-message"): (
        [],
        ["to_refs", "cc_refs", "bcc_refs"],
    ),
    ("email-message", None, "artifact"): (["raw_email_ref", "body_raw_ref"], []),
    ("email-message", None, "file"): (
        ["body_raw_ref"],
        [],
    ),  # FIXME: should be mime-part-type?
    # ip address
    ("autonomous-system", "owned", "ipv4-addr"): ([], ["belongs_to_refs"]),
    ("autonomous-system", "owned", "ipv6-addr"): ([], ["belongs_to_refs"]),
    # network-traffic
    ("ipv4-addr", "created", "network-traffic"): ([], ["src_ref"]),
    ("ipv6-addr", "created", "network-traffic"): ([], ["src_ref"]),
    ("mac-addr", "created", "network-traffic"): ([], ["src_ref"]),
    ("domain-name", "created", "network-traffic"): ([], ["src_ref"]),
    ("artifact", "created", "network-traffic"): ([], ["src_payload_ref"]),
    ("mac-addr", None, "ipv4-addr"): ([], ["resolves_to_refs"]),
    ("mac-addr", None, "ipv6-addr"): ([], ["resolves_to_refs"]),
    ("http-request-ext", None, "artifact"): (["message_body_data_ref"], []),
    ("ipv4-addr", "accepted", "network-traffic"): ([], ["dst_ref"]),
    ("ipv6-addr", "accepted", "network-traffic"): ([], ["dst_ref"]),
    ("mac-addr", "accepted", "network-traffic"): ([], ["dst_ref"]),
    ("domain-name", "accepted", "network-traffic"): ([], ["dst_ref"]),
    ("artifact", "accepted", "network-traffic"): ([], ["dst_payload_ref"]),
    ("network-traffic", "contained", "network-traffic"): (
        ["encapsulated_by_ref"],
        ["encapsulated_by_ref"],
    ),
    # process
    ("process", "created", "network-traffic"): (["opened_connection_refs"], []),
    ("user-account", "owned", "process"): ([], ["creator_user_ref"]),
    ("process", "loaded", "file"): (["binary_ref"], []),
    # ("process", "created", "process"): (["child_refs"], ["parent_ref"]),
    ("process", "created", "process"): ([], ["parent_ref"]),
    # service
    ("windows-service-ext", "loaded", "file"): (["service_dll_refs"], []),
    ("windows-service-ext", "loaded", "user-account"): (["creator_user_ref"], []),
}

# FIXME: is this no longer needed?
# the first available attribute will be used to uniquely identify the entity
stix_2_0_identical_mapping = {
    # entity-type: id attributes candidates
    "directory": ("path",),
    "domain-name": ("value",),
    "email-addr": ("value",),
    "file": ("name",),  # optional in STIX standard
    "ipv4-addr": ("value",),
    "ipv6-addr": ("value",),
    "mac-addr": ("value",),
    "mutex": ("name",),
    # `pid` is optional in STIX standard
    # `first_observed` cannot be used since it may be wrong (derived from observation)
    # `command_line` or `name` may not be in data and cannot be used
    "process": ("pid", "name"),
    "software": ("name",),
    "url": ("value",),
    "user-account": ("user_id",),  # optional in STIX standard
    "windows-registry-key": ("key",),  # optional in STIX standard
}

stix_x_ibm_event_mapping = {
    # entity-type to ref in x-oca-event
    "process": "process_ref",
    "domain-name": "domain_ref",
    "file": "file_ref",
    "user-account": "user_ref",
    "windows-registry-key": "registry_ref",
    "network-traffic": "nt_ref",
    "x-oca-asset": "host_ref",
}

# no direction for generic relations
generic_relations = ["linked"]

all_relations = list(
    set([x[1] for x in stix_2_0_ref_mapping.keys() if x[1]] + generic_relations)
)


def get_entity_id_attribute(variable):
    # this function should always return something
    # if no entity id attribute found, fall back to record "id" by default
    # this works for:
    #   - no appriparite identifier attribute found given specific data
    #   - "network-traffic" (not in stix_2_0_identical_mapping)
    id_attr = "id"

    if variable.type in stix_2_0_identical_mapping:
        available_attributes = variable.store.columns(variable.entity_table)
        for attr in stix_2_0_identical_mapping[variable.type]:
            if attr in available_attributes:
                query = Query()
                query.append(Table(variable.entity_table))
                query.append(Projection([attr]))
                query.append(Unique())
                rows = variable.store.run_query(query).fetchall()
                all_values = [row[attr] for row in rows if row[attr]]
                if all_values:
                    id_attr = attr
                    break

    return id_attr


def are_entities_associated_with_x_ibm_event(entity_types):
    flags = [entity_type in stix_x_ibm_event_mapping for entity_type in entity_types]
    return all(flags)


def compile_generic_relation_to_pattern(return_type, input_type, input_var_name):
    comp_exps = []
    for relation, is_reversed in _enumerate_relations_between_entities(
        return_type, input_type
    ):
        comp_exps += _generate_paramstix_comparison_expressions(
            return_type, relation, input_type, is_reversed, input_var_name
        )
    pattern = "[" + " OR ".join(comp_exps) + "]"
    _logger.debug(f"generic relation pattern compiled: {pattern}")
    return pattern


def compile_specific_relation_to_pattern(
    return_type, relation, input_type, is_reversed, input_var_name
):
    comp_exps = _generate_paramstix_comparison_expressions(
        return_type, relation, input_type, is_reversed, input_var_name
    )
    pattern = "[" + " OR ".join(comp_exps) + "]"
    _logger.debug(f"specific relation pattern compiled: {pattern}")
    return pattern


def compile_identical_entity_search_pattern(var_name, var_struct, does_support_id):
    # "id" attribute may not be available for STIX 2.0 via STIX-shifter
    # so `does_support_id` is set to False in default kestrel config file
    attribute = get_entity_id_attribute(var_struct)
    if attribute == "id" and not does_support_id:
        pattern_raw = None
    else:
        pattern_raw = f"[{var_struct.type}:{attribute} = {var_name}.{attribute}]"
    _logger.debug(f"identical entity search raw pattern generated: {pattern_raw}")
    return pattern_raw


def compile_x_ibm_event_search_flow_in_pattern(input_type, input_var_name):
    ref = stix_x_ibm_event_mapping[input_type]
    pattern = f"[x-oca-event:{ref}.id = {input_var_name}.id]"
    _logger.debug(f"x-oca-event flow in pattern compiled: {pattern}")
    return pattern


def compile_x_ibm_event_search_flow_out_pattern(return_type, input_event_var_name):
    ref = stix_x_ibm_event_mapping[return_type]
    pattern = f"[{return_type}:id = {input_event_var_name}.{ref}.id]"
    _logger.debug(f"x-oca-event flow out pattern compiled: {pattern}")
    return pattern


def _enumerate_relations_between_entities(return_type, input_type):
    # return: [(relation, is_reversed)]
    relations = []
    for (x, r, y) in stix_2_0_ref_mapping.keys():
        if x == return_type and y == input_type:
            relations.append((r, False))
        if y == return_type and x == input_type:
            relations.append((r, True))
    _logger.debug(
        f'enumerated relations between "{return_type}" and "{input_type}": {relations}'
    )
    return relations


def _generate_paramstix_comparison_expressions(
    return_type, relation, input_type, is_reversed, input_var_name
):
    (entity_x, entity_y) = (
        (input_type, return_type) if is_reversed else (return_type, input_type)
    )

    stix_src_refs, stix_tgt_refs = stix_2_0_ref_mapping[(entity_x, relation, entity_y)]

    comp_exps = []
    for stix_ref in stix_src_refs:
        if stix_ref.endswith("_refs"):
            comp_exps.append(f"{return_type}:id = {input_var_name}.{stix_ref}[*].id")
        else:
            comp_exps.append(f"{return_type}:id = {input_var_name}.{stix_ref}.id")

    for stix_ref in stix_tgt_refs:
        if stix_ref.endswith("_refs"):
            comp_exps.append(f"{return_type}:id = {input_var_name}.{stix_ref}[*].id")
        else:
            comp_exps.append(f"{return_type}:id = {input_var_name}.{stix_ref}.id")

    return comp_exps


def fine_grained_relational_process_filtering(
    local_var, prefetch_entity_table, store, config
):
    # Two-step search for matched processes
    # 1. anchor process search (find anchor process in pfeh_processes against ref_processes)
    # 2. precise process search (find process in pfeh_processes against anchor_processes)

    # two situations worth mentioning:
    # - in Linux, a new process will be forked, then exec to change name. In this case,
    #   we need to search for anchor_rows to identify process with even name changed,
    #   then get all records of both process before name change and after name change.
    # - ppid is useful to identify a process with pid, however, in one situation, the
    #   ppid data is not available in the first phase of FIND (creating of local_var
    #   using deref in firepit)---FIND parent process of current process. This is because
    #   most datasource does not store *parent parent process pid* for deref to get ppid
    #   of the parent. In this case, we need to search for anchor_rows to infer the ppid.

    _logger.debug(
        f"start fine-grained relational process filtering for prefetched table: {prefetch_entity_table}"
    )

    # reference processes obtained from de-referring data in firepit
    # type(ref_processes) == {pid: (rid, (pname, ppid, start_time, end_time))}
    ref_processes = _query_process_with_time_and_ppid(store, local_var.entity_table)

    # prefetched processes to be filtered
    # type(pfeh_processes) == {pid: (rid, (pname, ppid, start_time, end_time))}
    pfeh_processes = _query_process_with_time_and_ppid(store, prefetch_entity_table)

    # 1. anchor process search (a subset of pfeh_processes that matches ref_processes)
    # type(anchor_processes) == {pid: (rid, (pname, ppid, start_time, end_time))}
    anchor_processes = _search_for_potential_identical_process(
        ref_processes, pfeh_processes, config
    )

    anchor_proc_cnt = sum(map(len, anchor_processes.values()))
    prefetched_proc_cnt = sum(map(len, pfeh_processes.values()))
    _logger.debug(
        f"found {anchor_proc_cnt} anchor rows out of {prefetched_proc_cnt} raw prefetched."
    )

    # 2. precise process search (a larger subset of pfeh_processes that matches anchor_processes)
    # type(anchor_processes) == {pid: (rid, (pname, ppid, start_time, end_time))}
    filtered_processes = _search_for_potential_identical_process(
        anchor_processes, pfeh_processes, config
    )

    filtered_ids = [rid for procs in filtered_processes.values() for rid, _ in procs]
    filtered_ids = list(set(filtered_ids))

    _logger.debug(
        f"found {len(filtered_ids)} out of {prefetched_proc_cnt} raw prefetched results to be true relational process records."
    )

    return filtered_ids


def build_pattern_from_ids(return_type, ids):
    if ids:
        return (
            "[" + return_type + ":id IN (" + ", ".join(map(value_to_stix, ids)) + ")]"
        )
    else:
        return None


def _query_process_with_time_and_ppid(store, var_table_name):
    pid2procs = defaultdict(list)

    if "parent_ref" in store.columns(var_table_name):
        has_parent_ref = True
    else:
        has_parent_ref = False

    query_details = [
        Table(var_table_name),
        Join("__contains", "id", "=", "target_ref"),
        Join("observed-data", "source_ref", "=", "id"),
    ]

    if has_parent_ref:
        query_details.append(
            # put the LEFT JOIN at last, so no need to specify lhs for the first two JOINS
            Join(
                "process", "parent_ref", "=", "id", how="LEFT OUTER", lhs=var_table_name
            )
        )

    projection_details = [
        Column("id", var_table_name, "id"),
        Column("pid", var_table_name, "pid"),
        Column("name", var_table_name, "name"),
        "first_observed",
        "last_observed",
    ]

    if has_parent_ref:
        projection_details.append(Column("pid", "process", "ppid"))

    query_details.append(Projection(projection_details))

    query = Query(query_details)

    rows = store.run_query(query).fetchall()

    for row in rows:
        if row["pid"]:
            rid = row["id"]
            pname = row["name"]
            ppid = row["ppid"] if has_parent_ref else None
            st = dateutil.parser.isoparse(row["first_observed"])
            ed = dateutil.parser.isoparse(row["last_observed"])
            pid2procs[row["pid"]].append((rid, (pname, ppid, st, ed)))

    return pid2procs


def _search_for_potential_identical_process(ref_pid2procs, fil_pid2procs, config):
    # ref_pid2procs: {"pid":(rid, (proc_details))} for reference
    # fil_pid2procs: {"pid":(rid, (proc_details))} to search

    res_pid2procs = defaultdict(list)

    for (pid, fil_procs) in fil_pid2procs.items():
        for rid, fil_row in fil_procs:
            for _, ref_row in ref_pid2procs[pid]:
                if _identical_process_check(fil_row, ref_row, config):
                    res_pid2procs[pid].append((rid, fil_row))
                    break

    return res_pid2procs


def _identical_process_check(fil_row, ref_row, config):
    pbnc_bo = datetime.timedelta(
        seconds=config["pid_but_name_changed_time_begin_offset"]
    )
    pbnc_eo = datetime.timedelta(seconds=config["pid_but_name_changed_time_end_offset"])
    pan_bo = datetime.timedelta(seconds=config["pid_and_name_time_begin_offset"])
    pan_eo = datetime.timedelta(seconds=config["pid_and_name_time_end_offset"])
    pap_bo = datetime.timedelta(seconds=config["pid_and_ppid_time_begin_offset"])
    pap_eo = datetime.timedelta(seconds=config["pid_and_ppid_time_end_offset"])
    panap_bo = datetime.timedelta(
        seconds=config["pid_and_name_and_ppid_time_begin_offset"]
    )
    panap_eo = datetime.timedelta(
        seconds=config["pid_and_name_and_ppid_time_end_offset"]
    )

    fil_pname, fil_ppid, fil_start_time, fil_end_time = fil_row
    ref_pname, ref_ppid, ref_start_time, ref_end_time = ref_row
    if (
        (
            fil_pname
            and fil_ppid
            and fil_pname == ref_pname
            and fil_ppid == ref_ppid
            and fil_start_time > ref_start_time + panap_bo
            and fil_start_time < ref_end_time + panap_eo
        )
        or (
            fil_pname
            and fil_pname == ref_pname
            and fil_start_time > ref_start_time + pan_bo
            and fil_start_time < ref_end_time + pan_eo
        )
        or (
            fil_ppid
            and fil_ppid == ref_ppid
            and fil_start_time > ref_start_time + pap_bo
            and fil_start_time < ref_end_time + pap_eo
        )
        or (
            # name changed process, Linux fork+exec handled
            fil_start_time > ref_start_time + pbnc_bo
            and fil_start_time < ref_end_time + pbnc_eo
        )
        or (
            # name changed process, Linux fork+exec handled
            fil_end_time > ref_start_time + pbnc_bo
            and fil_end_time < ref_end_time + pbnc_eo
        )
    ):
        return True
    else:
        return False



from kestrel.absinterface import InterfaceManager
from kestrel.datasource import MODULE_PREFIX, AbstractDataSourceInterface
from kestrel.exceptions import (
    DataSourceInterfaceNotFound,
    InvalidDataSourceInterfaceImplementation,
    ConflictingDataSourceInterfaceScheme,
)


class DataSourceManager(InterfaceManager):
    def __init__(self, config):
        super().__init__(
            config,
            "datasources",
            ["language", "default_datasource_schema"],
            MODULE_PREFIX,
            AbstractDataSourceInterface,
            DataSourceInterfaceNotFound,
            InvalidDataSourceInterfaceImplementation,
            ConflictingDataSourceInterfaceScheme,
        )

        # important state keeper, required in Session()
        self.queried_data_sources = [None]

    def list_data_sources_from_scheme(self, scheme):
        i, c = self._get_interface_with_config(scheme)
        return i.list_data_sources(c)

    def query(self, uri, pattern, session_id):
        scheme, uri = self._parse_and_complete_uri(uri)
        i, c = self._get_interface_with_config(scheme)
        rs = i.query(uri, pattern, session_id, c)
        self.queried_data_sources.append(uri)
        return rs



"""The STIX-shifter data source package provides access to data sources via
`stix-shifter`_.

The STIX-shifter interface can reach multiple data sources. The user needs to
provide one *profile* per data source. The profile name (case insensitive) will
be used in the ``FROM`` clause of the Kestrel ``GET`` command, e.g., ``newvar =
GET entity-type FROM stixshifter://profilename WHERE ...``. Kestrel runtime
will load profiles from 3 places (the later will override the former):

#. STIX-shifter interface config file (only when a Kestrel session starts):

    Put your profiles in the STIX-shifter interface config file (YAML):

    - Default path: ``~/.config/kestrel/stixshifter.yaml``.
    - A customized path specified in the environment variable ``KESTREL_STIXSHIFTER_CONFIG``.

    Example of STIX-shifter interface config file containing profiles:

    .. code-block:: yaml

        profiles:
            host101:
                connector: elastic_ecs
                connection:
                    host: elastic.securitylog.company.com
                    port: 9200
                    selfSignedCert: false # this means do NOT check cert
                    indices: host101
                config:
                    auth:
                        id: VuaCfGcBCdbkQm-e5aOx
                        api_key: ui2lp2axTNmsyakw9tvNnw
            host102:
                connector: qradar
                connection:
                    host: qradar.securitylog.company.com
                    port: 443
                config:
                    auth:
                        SEC: 123e4567-e89b-12d3-a456-426614174000
            host103:
                connector: cbcloud
                connection:
                    host: cbcloud.securitylog.company.com
                    port: 443
                config:
                    auth:
                        org-key: D5DQRHQP
                        token: HT8EMI32DSIMAQ7DJM

#. environment variables (only when a Kestrel session starts):

    Three environment variables are required for each profile:

    - ``STIXSHIFTER_PROFILENAME_CONNECTOR``: the STIX-shifter connector name,
      e.g., ``elastic_ecs``.
    - ``STIXSHIFTER_PROFILENAME_CONNECTION``: the STIX-shifter `connection
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#connection>`_
      object in JSON string.
    - ``STIXSHIFTER_PROFILENAME_CONFIG``: the STIX-shifter `configuration
      <https://github.com/opencybersecurityalliance/stix-shifter/blob/master/OVERVIEW.md#configuration>`_
      object in JSON string.

    Example of environment variables for a profile:

    .. code-block:: console

        $ export STIXSHIFTER_HOST101_CONNECTOR=elastic_ecs
        $ export STIXSHIFTER_HOST101_CONNECTION='{"host":"elastic.securitylog.company.com", "port":9200, "indices":"host101"}'
        $ export STIXSHIFTER_HOST101_CONFIG='{"auth":{"id":"VuaCfGcBCdbkQm-e5aOx", "api_key":"ui2lp2axTNmsyakw9tvNnw"}}'

#. any in-session edit through the ``CONFIG`` command.

If you launch Kestrel in debug mode, STIX-shifter debug mode is still not
enabled by default. To record debug level logs of STIX-shifter, create
environment variable ``KESTREL_STIXSHIFTER_DEBUG`` with any value.

.. _STIX-shifter: https://github.com/opencybersecurityalliance/stix-shifter

"""

import json
import time
import copy
import logging

from stix_shifter.stix_translation import stix_translation
from stix_shifter.stix_transmission import stix_transmission

from kestrel.utils import mkdtemp
from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import DataSourceError, DataSourceManagerInternalError
from kestrel_datasource_stixshifter.connector import check_module_availability
from kestrel_datasource_stixshifter.config import (
    RETRIEVAL_BATCH_SIZE,
    get_datasource_from_profiles,
    load_profiles,
    set_stixshifter_logging_level,
)

_logger = logging.getLogger(__name__)


class StixShifterInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        """STIX-shifter data source interface only supports ``stixshifter://`` scheme."""
        return ["stixshifter"]

    @staticmethod
    def list_data_sources(config):
        """Get configured data sources from environment variable profiles."""

        # CONFIG command is not supported
        # profiles will be updated according to YAML file and env var
        config["profiles"] = load_profiles()

        data_sources = list(config["profiles"].keys())
        data_sources.sort()
        return data_sources

    @staticmethod
    def query(uri, pattern, session_id, config):
        """Query a stixshifter data source."""

        # CONFIG command is not supported
        # profiles will be updated according to YAML file and env var
        config["profiles"] = load_profiles()

        scheme, _, profile = uri.rpartition("://")
        profiles = profile.split(",")
        if scheme != "stixshifter":
            raise DataSourceManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        set_stixshifter_logging_level()

        ingestdir = mkdtemp()
        query_id = ingestdir.name
        bundles = []
        _logger.debug(f"prepare query with ID: {query_id}")
        for i, profile in enumerate(profiles):
            # STIX-shifter will alter the config objects, thus making them not reusable.
            # So only give STIX-shifter a copy of the configs.
            # Check `modernize` functions in the `stix_shifter_utils` for details.
            (connector_name, connection_dict, configuration_dict,) = map(
                copy.deepcopy, get_datasource_from_profiles(profile, config["profiles"])
            )

            check_module_availability(connector_name)

            data_path_striped = "".join(filter(str.isalnum, profile))
            ingestfile = ingestdir / f"{i}_{data_path_striped}.json"

            query_metadata = json.dumps(
                {"id": "identity--" + query_id, "name": connector_name}
            )

            translation = stix_translation.StixTranslation()
            transmission = stix_transmission.StixTransmission(
                connector_name, connection_dict, configuration_dict
            )

            translation_options = copy.deepcopy(connection_dict.get("options", {}))
            dsl = translation.translate(
                connector_name, "query", query_metadata, pattern, translation_options
            )

            if "error" in dsl:
                raise DataSourceError(
                    f"STIX-shifter translation from STIX to native query failed with message: {dsl['error']}"
                )

            _logger.debug(f"STIX pattern to query: {pattern}")
            _logger.debug(f"translate results: {dsl}")

            # query results should be put together; when translated to STIX, the relation between them will remain
            connector_results = []
            for query in dsl["queries"]:
                search_meta_result = transmission.query(query)
                if search_meta_result["success"]:
                    search_id = search_meta_result["search_id"]
                    if transmission.is_async():
                        time.sleep(1)
                        status = transmission.status(search_id)
                        if status["success"]:
                            while (
                                status["progress"] < 100
                                and status["status"] == "RUNNING"
                            ):
                                status = transmission.status(search_id)
                        else:
                            stix_shifter_error_msg = (
                                status["error"]
                                if "error" in status
                                else "details not avaliable"
                            )
                            raise DataSourceError(
                                f"STIX-shifter transmission.status() failed with message: {stix_shifter_error_msg}"
                            )

                    result_retrieval_offset = 0
                    has_remaining_results = True
                    while has_remaining_results:
                        result_batch = transmission.results(
                            search_id, result_retrieval_offset, RETRIEVAL_BATCH_SIZE
                        )
                        if result_batch["success"]:
                            new_entries = result_batch["data"]
                            if new_entries:
                                connector_results += new_entries
                                result_retrieval_offset += RETRIEVAL_BATCH_SIZE
                                if len(new_entries) < RETRIEVAL_BATCH_SIZE:
                                    has_remaining_results = False
                            else:
                                has_remaining_results = False
                        else:
                            stix_shifter_error_msg = (
                                result_batch["error"]
                                if "error" in result_batch
                                else "details not avaliable"
                            )
                            raise DataSourceError(
                                f"STIX-shifter transmission.results() failed with message: {stix_shifter_error_msg}"
                            )

                else:
                    stix_shifter_error_msg = (
                        search_meta_result["error"]
                        if "error" in search_meta_result
                        else "details not avaliable"
                    )
                    raise DataSourceError(
                        f"STIX-shifter transmission.query() failed with message: {stix_shifter_error_msg}"
                    )

            _logger.debug("transmission succeeded, start translate back to STIX")

            stixbundle = translation.translate(
                connector_name,
                "results",
                query_metadata,
                json.dumps(connector_results),
                translation_options,
            )

            if "error" in stixbundle:
                raise DataSourceError(
                    f"STIX-shifter translation results to STIX failed with message: {stixbundle['error']}"
                )

            _logger.debug(f"dumping STIX bundles into file: {ingestfile}")
            with ingestfile.open("w") as ingest:
                json.dump(stixbundle, ingest, indent=4)
            bundles.append(str(ingestfile.expanduser().resolve()))

        return ReturnFromFile(query_id, bundles)



from kestrel.exceptions import VariableNotExist


class SymbolTable(dict):
    def __getitem__(self, var_name):
        try:
            return super().__getitem__(var_name)
        except KeyError as e:
            raise VariableNotExist(var_name)



################################################################
#                     Generic Errors
################################################################


class KestrelException(Exception):
    """Generic Kestrel Exception

    Args:
        error (str): error message.
        suggestion (str): suggestion to fix the issue.
    """

    def __init__(self, error, suggestion=""):
        self.error = error[:-1] if error[-1] == "\n" else error

        self.suggestion = (
            suggestion
            if (not suggestion or suggestion[-1] == ".")
            else suggestion + "."
        )

    def __str__(self):
        return f"[ERROR] {self.__class__.__name__}: {self.error}\n{self.suggestion}"


class KestrelInternalError(KestrelException):
    def __init__(self, error):
        super().__init__(error, "please open a github issue to report")


class KestrelNotImplemented(KestrelException):
    def __init__(self, error):
        super().__init__(
            "Functionality not implemented: " + error,
            "please search for the github issue to add comment",
        )


################################################################
#                     Kestrel Session Errors
################################################################


class InvalidConfiguration(KestrelException):
    def __init__(self, error, suggestion):
        super().__init__(error, suggestion)


class DebugCacheLinkOccupied(KestrelException):
    def __init__(self, link):
        super().__init__(
            f"the path '{link}' is used by other users and cannot be removed (permission error).",
            "Try manually remove the file/link or set 'debug.cache_directory_prefix' in the Kestrel config file to other values",
        )


################################################################
#                       Kestrel Syntax Errors
################################################################


class KestrelSyntaxError(KestrelException):
    def __init__(self, line, column, invalid_term_type, invalid_term_value, expected):
        self.line = line
        self.column = column
        self.invalid_term_type = invalid_term_type
        self.invalid_term_value = invalid_term_value
        self.expected = list(expected)
        self._expects_str = (
            f'expects "{self.expected[0]}"'
            if len(self.expected) == 1
            else f"expects one of {self.expected}"
        )
        super().__init__(
            f'invalid {self.invalid_term_type} "{self.invalid_term_value}" at line {self.line} column {self.column}, {self._expects_str}',
            "rewrite the failed statement",
        )


class InvalidStixPattern(KestrelException):
    def __init__(
        self,
        stix,
        line=None,
        column=None,
        invalid_term_type=None,
        invalid_term_value=None,
    ):
        self.stix = stix
        self.line = line
        self.column = column
        self.invalid_term_type = invalid_term_type
        self.invalid_term_value = invalid_term_value
        msg = f'invalid STIX pattern "{stix}"'
        if self.invalid_term_value:
            details = f': invalid {self.invalid_term_type} "{self.invalid_term_value}" at line {self.line} column {self.column}'
            msg = msg + details
        super().__init__(msg, "rewrite the STIX pattern")


class InvalidECGPattern(KestrelException):
    pass


class MissingDataSource(KestrelException):
    def __init__(self, stmt):
        super().__init__(f"missing datasource in statement: {str(stmt)}")


class VariableNotExist(KestrelException):
    def __init__(self, var_name):
        self.var_name = var_name
        super().__init__(
            f'variable "{var_name}" does not exist', "check the variable used"
        )


class UnsupportedRelation(KestrelException):
    def __init__(self, entity_x, relation, entity_y):
        super().__init__(
            f'unsupported relation "{entity_x}--{relation}--{entity_y}"',
            "check for supported relations and entity types in the documentation",
        )


################################################################
#                 Kestrel Code Generation Errors
################################################################


class EmptyInputVariable(KestrelException):
    def __init__(self, var_name):
        super().__init__(
            f'empty input variable "{var_name}"',
            "rewrite the command to use a non-empty variable as input",
        )


class InvalidAttribute(KestrelException):
    def __init__(self, attribute):
        super().__init__(
            f'invalid attribute "{attribute}"',
            "rewrite the command with a valid attribute",
        )


class NonUniformEntityType(KestrelException):
    def __init__(self, etypes):
        super().__init__(
            f"there are more than one entity types in input data: {etypes}",
            "provide homogeneous entities to construct a Kestrel variable",
        )


class MissingEntityType(KestrelException):
    def __init__(self):
        super().__init__(
            'input data does not have "type" column',
            'add "type" column to data or specify entity type in the Kestrel command',
        )


class MissingEntityAttribute(KestrelException):
    def __init__(self, var_name, attribute):
        super().__init__(
            f'variable "{var_name}" does not have required attribute "{attribute}"',
            "remove transform or specify different variable in the Kestrel command",
        )


################################################################
#                     Data Source Errors
################################################################


class DataSourceConnectionError(KestrelException):
    def __init__(self, uri):
        super().__init__(
            f"cannot establish connection to {uri}",
            "check URI for typos; please test network connection",
        )


class DataSourceManagerInternalError(KestrelInternalError):
    def __init__(self, error):
        super().__init__(error)


class InvalidDataSource(KestrelException):
    def __init__(self, uri, itf, msg=""):
        super().__init__(
            f'invalid data source "{uri}" at interface "{itf}". {msg}',
            "please check data source configuration",
        )


class DataSourceError(KestrelException):
    def __init__(self, error, suggestion=""):
        if not suggestion:
            suggestion = "please check data source config or test the query manually"
        super().__init__(
            error,
            suggestion,
        )


class DataSourceInterfaceNotFound(KestrelException):
    def __init__(self, scheme):
        super().__init__(
            f'interface handling "{scheme}://" does not exist',
            "(re)install the missing/broken data source interface package",
        )


class InvalidDataSourceInterfaceImplementation(Exception):
    def __init__(self, error):
        super().__init__(error, "report to data source interface developer")


class ConflictingDataSourceInterfaceScheme(KestrelException):
    def __init__(self, itf_a, itf_b, scheme):
        super().__init__(
            f'conflicting data source interface scheme "{scheme}" between "{itf_a.__module__}" and "{itf_b.__module__}"',
            "uninstall one of the data source interfaces",
        )


################################################################
#                       Analytics Errors
################################################################


class AnalyticsManagerInternalError(KestrelInternalError):
    def __init__(self, error):
        super().__init__(error)


class InvalidAnalytics(KestrelException):
    def __init__(self, name, itf, msg=""):
        super().__init__(
            f'invalid analytics "{name}" at interface "{itf}". {msg}',
            "please check analytics availability",
        )


class AnalyticsError(KestrelException):
    def __init__(self, error, suggestion=""):
        suggestion = "report to analytics developer" if not suggestion else suggestion
        super().__init__(error, suggestion)


class AnalyticsInterfaceNotFound(KestrelException):
    def __init__(self, scheme):
        super().__init__(
            f'interface handling "{scheme}://" does not exist',
            "(re)install the missing/broken data source interface package",
        )


class InvalidAnalyticsInterfaceImplementation(Exception):
    def __init__(self, error):
        super().__init__(error, "report to analytics interface developer")


class ConflictingAnalyticsInterfaceScheme(KestrelException):
    def __init__(self, itf_a, itf_b, scheme):
        super().__init__(
            f'conflicting analytics interface scheme "{scheme}" between "{itf_a.__module__}" and "{itf_b.__module__}"',
            "uninstall one of the analytics interfaces",
        )


class InvalidAnalyticsArgumentCount(KestrelException):
    def __init__(self, analytics_name, num_received, num_expected):
        super().__init__(
            f'the analytics "{analytics_name}" takes {num_expected} Kestrel variables, not {num_received} as given in APPLY.'
        )


class InvalidAnalyticsInput(KestrelException):
    def __init__(self, type_received, types_expected):
        typelist = ", ".join([f'"{t}"' for t in types_expected])
        super().__init__(
            f'received unsupported type "{type_received}"; expected one of {typelist}'
        )


class InvalidAnalyticsOutput(KestrelException):
    def __init__(self, analytics_name, return_type):
        super().__init__(
            f"unsupported return type {return_type} from analytics: {analytics_name}"
        )



################################################################
#                       Module Summary
#
# - Code generation for each command in kestrel.lark
#   - The execution function names match commands in kestrel.lark
# - Each command takes 2 arguments
#     ( statement, session )
#   - statement is the current statement to process,
#     which is a dict from the parser
#   - session is the current session (context)
# - Every command returns a tuple (VarStruct, Display)
#   - VarStruct is a new object associated with the output var
#     - VarStruct associated with stmt["output"]
#     - None for some commands, e.g., DISP, SAVE, STAT
#   - Display is the data to display on the user interface
#     - a string
#     - a list of (str,str|list(str)) tuples
#     - a table that can be imported to pandas dataframe
################################################################

import functools
import logging
import itertools
from collections import OrderedDict
from copy import deepcopy

from firepit.deref import auto_deref
from firepit.exceptions import InvalidAttr
from firepit.query import Limit, Offset, Order, Projection, Query
from firepit.stix20 import summarize_pattern

from kestrel.semantics.reference import make_deref_func, make_var_timerange_func
from kestrel.utils import remove_empty_dicts, dedup_ordered_dicts, lowered_str_list
from kestrel.exceptions import *
from kestrel.symboltable.variable import new_var
from kestrel.symboltable.symtable import SymbolTable
from kestrel.syntax.parser import parse_ecgpattern
from kestrel.syntax.utils import (
    get_entity_types,
    get_all_input_var_names,
    timedelta_seconds,
)
from kestrel.codegen.data import load_data, load_data_file, dump_data_to_file
from kestrel.codegen.display import DisplayDataframe, DisplayDict, DisplayWarning
from kestrel.codegen.queries import (
    compile_specific_relation_to_query,
    compile_generic_relation_to_query,
)
from kestrel.codegen.relations import (
    generic_relations,
    compile_identical_entity_search_pattern,
    fine_grained_relational_process_filtering,
    get_entity_id_attribute,
    stix_2_0_identical_mapping,
    build_pattern_from_ids,
)

_logger = logging.getLogger(__name__)


################################################################
#                       Private Decorators
################################################################


def _default_output(func):
    # by default, create a table/view in the backend
    # using the output var name
    # in this case, the store backend can return no VarStruct
    @functools.wraps(func)
    def wrapper(stmt, session):
        ret = func(stmt, session)
        if not ret:
            var_struct = new_var(
                session.store, stmt["output"], [], stmt, session.symtable
            )
            return var_struct, None
        else:
            return ret

    return wrapper


def _guard_empty_input(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        for varname in get_all_input_var_names(stmt):
            v = session.symtable[varname]
            if v.length + v.records_count == 0:
                raise EmptyInputVariable(varname)
        else:
            return func(stmt, session)

    return wrapper


def _debug_logger(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        _logger.debug(f"Executing '{func.__name__}' with statement: {stmt}")
        return func(stmt, session)

    return wrapper


################################################################
#                 Code Generation for Commands
################################################################


@_debug_logger
@_default_output
def assign(stmt, session):
    entity_table = session.symtable[stmt["input"]].entity_table
    transform = stmt.get("transform")
    if transform:
        if transform.lower() == "timestamped":
            qry = session.store.timestamped(entity_table, run=False)
        else:
            qry = Query(entity_table)
    else:
        qry = Query(entity_table)

    qry = _build_query(session.store, entity_table, qry, stmt)

    try:
        session.store.assign_query(stmt["output"], qry)
        output = new_var(session.store, stmt["output"], [], stmt, session.symtable)
    except InvalidAttr as e:
        var_attr = str(e).split()[-1]
        var_name, _, attr = var_attr.rpartition(".")
        raise MissingEntityAttribute(var_name, attr) from e

    return output, None


@_debug_logger
@_default_output
def merge(stmt, session):
    entity_types = list(
        set([session.symtable[var_name].type for var_name in stmt["inputs"]])
    )
    if len(entity_types) > 1:
        raise NonUniformEntityType(entity_types)
    entity_tables = [
        session.symtable[var_name].entity_table for var_name in stmt["inputs"]
    ]
    session.store.merge(stmt["output"], entity_tables)
    output = new_var(session.store, stmt["output"], [], stmt, session.symtable)
    return output, None


@_debug_logger
@_default_output
def new(stmt, session):
    stmt["type"] = load_data(session.store, stmt["output"], stmt["data"], stmt["type"])


@_debug_logger
@_default_output
def load(stmt, session):
    stmt["type"] = load_data_file(
        session.store, stmt["output"], stmt["path"], stmt["type"]
    )


@_debug_logger
@_guard_empty_input
def save(stmt, session):
    dump_data_to_file(
        session.store, session.symtable[stmt["input"]].entity_table, stmt["path"]
    )
    return None, None


@_debug_logger
def info(stmt, session):
    header = session.store.columns(session.symtable[stmt["input"]].entity_table)
    direct_attrs, associ_attrs, custom_attrs, references = [], [], [], []
    for field in header:
        if field.startswith("x_"):
            custom_attrs.append(field)
        elif (
            field.endswith("_ref")
            or field.endswith("_refs")
            or field.endswith("_reference")
            or field.endswith("_references")
        ):
            # not useful in existing version, so do not display
            references.append(field)
        elif "_ref." in field or "_ref_" in field:
            associ_attrs.append(field)
        else:
            direct_attrs.append(field)

    disp = OrderedDict()
    disp["Entity Type"] = session.symtable[stmt["input"]].type
    disp["Number of Entities"] = str(len(session.symtable[stmt["input"]]))
    disp["Number of Records"] = str(session.symtable[stmt["input"]].records_count)
    disp["Entity Attributes"] = ", ".join(direct_attrs)
    disp["Indirect Attributes"] = [
        ", ".join(g)
        for _, g in itertools.groupby(associ_attrs, lambda x: x.rsplit(".", 1)[0])
    ]
    disp["Customized Attributes"] = ", ".join(custom_attrs)
    disp["Birth Command"] = session.symtable[stmt["input"]].birth_statement["command"]
    disp["Associated Datasource"] = session.symtable[stmt["input"]].data_source
    disp["Dependent Variables"] = ", ".join(
        session.symtable[stmt["input"]].dependent_variables
    )

    return None, DisplayDict(disp)


@_debug_logger
def disp(stmt, session):
    entity_table = session.symtable[stmt["input"]].entity_table
    transform = stmt.get("transform")
    if transform and entity_table:
        if transform.lower() == "timestamped":
            qry = session.store.timestamped(entity_table, run=False)
        else:
            qry = Query(entity_table)
    else:
        qry = Query(entity_table)

    qry = _build_query(session.store, entity_table, qry, stmt)
    try:
        cursor = session.store.run_query(qry)
    except InvalidAttr as e:
        var_attr = str(e).split()[-1]
        var_name, _, attr = var_attr.rpartition(".")
        raise MissingEntityAttribute(var_name, attr) from e
    content = cursor.fetchall()

    return None, DisplayDataframe(dedup_ordered_dicts(remove_empty_dicts(content)))


@_debug_logger
@_default_output
def get(stmt, session):
    pattern = stmt["stixpattern"]
    local_var_table = stmt["output"] + "_local"
    return_var_table = stmt["output"]
    return_type = stmt["type"]
    start_offset = session.config["stixquery"]["timerange_start_offset"]
    end_offset = session.config["stixquery"]["timerange_stop_offset"]
    display = None

    if "variablesource" in stmt:
        input_type = session.symtable[stmt["variablesource"]].type
        output_type = stmt["type"]
        if input_type != output_type:
            raise InvalidECGPattern(
                f"input variable type {input_type} does not match output type {output_type}"
            )
        session.store.filter(
            stmt["output"],
            stmt["type"],
            input_type,
            pattern,
        )
        output = new_var(session.store, return_var_table, [], stmt, session.symtable)
        _logger.debug(f"get from variable source \"{stmt['variablesource']}\"")

    elif "datasource" in stmt:
        # rs: RetStruct
        rs = session.data_source_manager.query(
            stmt["datasource"], pattern, session.session_id
        )
        query_id = rs.load_to_store(session.store)
        session.store.extract(local_var_table, return_type, query_id, pattern)
        _output = new_var(session.store, local_var_table, [], stmt, session.symtable)
        _logger.debug(
            f"native GET pattern executed and DB view {local_var_table} extracted."
        )

        pat_summary = summarize_pattern(pattern)
        pat_types = list(pat_summary.keys())
        if return_type in stix_2_0_identical_mapping:
            id_attrs = set(stix_2_0_identical_mapping[return_type])
        else:
            id_attrs = pat_summary[return_type]  # Hack
        if (
            len(pat_types) == 1
            and pat_types[0] == return_type
            and pat_summary[return_type] == id_attrs
        ):
            # Prefetch won't return anything new here, so skip it
            _logger.debug("To skip prefetch for direct query")
            is_direct_query = True
        else:
            is_direct_query = False

        if (
            not is_direct_query
            and _is_prefetch_allowed_in_config(
                session.config["prefetch"], "get", return_type
            )
            and len(_output)
        ):
            prefetch_ret_var_table = return_var_table + "_prefetch"
            ext_graph_pattern = deepcopy(stmt["where"])
            ext_graph_pattern.prune_away_centered_graph(return_type)
            prefetch_ret_entity_table = _prefetch(
                return_type,
                prefetch_ret_var_table,
                local_var_table,
                stmt["timerange"],
                start_offset,
                end_offset,
                SymbolTable({local_var_table: _output}),
                session.store,
                ext_graph_pattern,
                session.session_id,
                session.data_source_manager,
                session.config["stixquery"]["support_id"],
            )

            if return_type == "process" and get_entity_id_attribute(_output) != "id":
                prefetch_ret_entity_table = _filter_prefetched_process(
                    return_var_table,
                    session,
                    _output,
                    prefetch_ret_entity_table,
                    return_type,
                )
        else:
            prefetch_ret_entity_table = None

        if prefetch_ret_entity_table:
            _logger.debug(
                f"merge {local_var_table} and {prefetch_ret_entity_table} into {return_var_table}."
            )
            session.store.merge(
                return_var_table, [local_var_table, prefetch_ret_entity_table]
            )
            for v in list(
                set(
                    [local_var_table, prefetch_ret_entity_table, prefetch_ret_var_table]
                )
            ):
                if not session.debug_mode:
                    _logger.debug(f"remove temp store view {v}.")
                    session.store.remove_view(v)
        else:
            _logger.debug(
                f'prefetch return None, just rename native GET pattern matching results into "{return_var_table}".'
            )
            session.store.rename_view(local_var_table, return_var_table)

        output = new_var(session.store, return_var_table, [], stmt, session.symtable)

        if not len(output):
            if not return_type.startswith("x-") and return_type not in (
                set(session.store.types()) | set(get_entity_types())
            ):
                display = DisplayWarning(f'unknown entity type "{return_type}"')
    else:
        raise KestrelInternalError(f"unknown type of source in {str(stmt)}")

    return output, display


@_debug_logger
@_default_output
@_guard_empty_input
def find(stmt, session):
    return_type = stmt["type"]
    input_type = session.symtable[stmt["input"]].type
    input_var_name = stmt["input"]
    return_var_table = stmt["output"]
    local_var_table = stmt["output"] + "_local"
    relation = stmt["relation"]
    is_reversed = stmt["reversed"]
    start_offset = session.config["stixquery"]["timerange_start_offset"]
    end_offset = session.config["stixquery"]["timerange_stop_offset"]
    rel_query = None

    if return_type not in session.store.types():
        # return empty variable
        output = new_var(session.store, None, [], stmt, session.symtable)

    else:
        _symtable = SymbolTable({input_var_name: session.symtable[input_var_name]})
        input_var_attrs = session.store.columns(input_type)
        return_type_attrs = session.store.columns(return_type)

        # First, get information from local store
        if relation in generic_relations:
            rel_query = compile_generic_relation_to_query(
                return_type, input_type, session.symtable[input_var_name].entity_table
            )

        else:
            rel_query = compile_specific_relation_to_query(
                return_type,
                relation,
                input_type,
                is_reversed,
                input_var_name,
                input_var_attrs,
                return_type_attrs,
            )

        # `session.store.assign_query` will generate new entity_table named `local_var_table`
        if rel_query:
            session.store.assign_query(local_var_table, rel_query, return_type)
            _output = new_var(
                session.store, local_var_table, [], stmt, session.symtable
            )

            # Second, prefetch all records of the entities and associated entities
            if (
                _is_prefetch_allowed_in_config(
                    session.config["prefetch"], "find", return_type
                )
                and len(_output)
                and _output.data_source
            ):
                prefetch_ret_var_table = return_var_table + "_prefetch"
                ext_graph_pattern = stmt["where"] if "where" in stmt else None
                prefetch_ret_entity_table = _prefetch(
                    return_type,
                    prefetch_ret_var_table,
                    local_var_table,
                    stmt["timerange"],
                    start_offset,
                    end_offset,
                    SymbolTable({local_var_table: _output}),
                    session.store,
                    ext_graph_pattern,
                    session.session_id,
                    session.data_source_manager,
                    session.config["stixquery"]["support_id"],
                )

                # special handling for process to filter out impossible relational processes
                # this is needed since STIX 2.0 does not have mandatory fields for
                # process and field like `pid` is not unique
                if (
                    return_type == "process"
                    and get_entity_id_attribute(_output) != "id"
                ):
                    prefetch_ret_entity_table = _filter_prefetched_process(
                        return_var_table,
                        session,
                        _output,
                        prefetch_ret_entity_table,
                        return_type,
                    )
            else:
                prefetch_ret_entity_table = None

            if prefetch_ret_entity_table:
                _logger.debug(
                    f"merge {local_var_table} and {prefetch_ret_entity_table} into {return_var_table}."
                )
                session.store.merge(
                    return_var_table, [local_var_table, prefetch_ret_entity_table]
                )
                for v in list(
                    set(
                        [
                            local_var_table,
                            prefetch_ret_entity_table,
                            prefetch_ret_var_table,
                        ]
                    )
                ):
                    if not session.debug_mode:
                        _logger.debug(f"remove temp store view {v}.")
                        session.store.remove_view(v)
            else:
                _logger.debug(
                    f'prefetch return None, just rename native GET pattern matching results into "{return_var_table}".'
                )
                session.store.rename_view(local_var_table, return_var_table)

        else:
            return_var_table = None
            _logger.info(f'no relation "{relation}" on this dataset')

        output = new_var(session.store, return_var_table, [], stmt, session.symtable)

    return output, None


@_debug_logger
@_default_output
@_guard_empty_input
def join(stmt, session):
    session.store.join(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        stmt["attribute_1"],
        session.symtable[stmt["input_2"]].entity_table,
        stmt["attribute_2"],
    )


@_debug_logger
@_default_output
@_guard_empty_input
def group(stmt, session):
    if "aggregations" in stmt:
        aggs = [(i["func"], i["attr"], i["alias"]) for i in stmt["aggregations"]]
    else:
        aggs = None
    session.store.group(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        stmt["attributes"],
        aggs,
    )


@_debug_logger
@_default_output
@_guard_empty_input
def sort(stmt, session):
    session.store.assign(
        stmt["output"],
        session.symtable[stmt["input"]].entity_table,
        op="sort",
        by=stmt["attribute"],
        ascending=stmt["ascending"],
    )


@_debug_logger
@_default_output
@_guard_empty_input
def apply(stmt, session):
    arg_vars = [session.symtable[v_name] for v_name in stmt["inputs"]]
    display = session.analytics_manager.execute(
        stmt["analytics_uri"], arg_vars, session.session_id, stmt["arguments"]
    )
    return None, display


################################################################
#                       Helper Functions
################################################################


def _prefetch(
    return_type,
    return_var_name,
    input_var_name,
    time_range,
    start_offset,
    end_offset,
    symtable,
    store,
    where_clause,
    session_id,
    ds_manager,
    does_support_id,
):
    """prefetch identical entities and associated entities.

    Put the input entities in the center of an observation and query the remote
    data source of associated with input variable, so we get back:

    1. all records about the input entities.

    2. associated entities such as parent/child processes of processes, processes of network-traffic, etc.

    The function does not have explicit return, but a side effect: a view in
    the store named after `return_var_name`.

    Args:
        input_var_name (str): input variable name.

        return_var_name (str): return variable name.

        return_type (str): return entity type.

        time_range ((datetime, datetime)).

        start_offset (int): start time offset by seconds.

        end_offset (int): end time offset by seconds.

        symtable ({str:VarStruct}): should has ``input_var_name``.

        store (firepit.SqlStorage): store.

        where_clause (ExtCenteredGraphPattern): pattern to merge to the prefetch auto-gen pattern

        session_id (str): session ID.

        does_support_id (bool): whether "id" can be an attribute in data source query.

    Returns:
        str: the entity table in store if the prefetch is performed else None.
    """

    _logger.debug(f"prefetch {return_type} to extend {input_var_name}.")

    pattern_raw = compile_identical_entity_search_pattern(
        input_var_name, symtable[input_var_name], does_support_id
    )

    if pattern_raw:

        pattern_ast = parse_ecgpattern(pattern_raw)
        deref_func = make_deref_func(store, symtable)
        get_timerange_func = make_var_timerange_func(store, symtable)
        pattern_ast.deref(deref_func, get_timerange_func)
        pattern_ast.add_center_entity(symtable[input_var_name].type)
        time_adj = tuple(map(timedelta_seconds, (start_offset, end_offset)))
        _logger.info(f"ext pattern in prefetch: {where_clause}")
        _logger.info(f"prefetch pattern before extend: {pattern_ast}")
        pattern_ast.extend("AND", where_clause)
        _logger.info(f"prefetch pattern after extend: {pattern_ast}")
        stix_pattern = pattern_ast.to_stix(time_range, time_adj)
        _logger.info(f"STIX pattern generated in prefetch: {stix_pattern}")

        if stix_pattern:
            data_source = symtable[input_var_name].data_source
            resp = ds_manager.query(data_source, stix_pattern, session_id)
            query_id = resp.load_to_store(store)

            # build the return_var_name view in store
            store.extract(return_var_name, return_type, query_id, stix_pattern)

            _logger.debug(f"prefetch successful.")
            return return_var_name

    _logger.info(f"prefetch return empty.")
    return None


def _filter_prefetched_process(
    return_var_name, session, local_var, prefetched_entity_table, return_type
):

    _logger.debug(f"filter prefetched {return_type} for {prefetched_entity_table}.")

    prefetch_filtered_var_name = return_var_name + "_prefetch_filtered"
    entity_ids = fine_grained_relational_process_filtering(
        local_var,
        prefetched_entity_table,
        session.store,
        session.config["prefetch"]["process_identification"],
    )
    if entity_ids:
        id_pattern = build_pattern_from_ids(return_type, entity_ids)
        session.store.extract(prefetch_filtered_var_name, return_type, None, id_pattern)
        _logger.debug("filter successful.")
        return prefetch_filtered_var_name
    else:
        _logger.info("no prefetched process found after filtering.")
        return None


def _set_projection(store, entity_table, query, paths):
    joins, proj = auto_deref(store, entity_table, paths=paths)
    query.joins.extend(joins)
    if query.proj:
        # Need to merge projections?  More-specific overrides less-specific ("*")
        new_cols = []
        for p in query.proj.cols:
            if not (hasattr(p, "table") and p.table == entity_table and p.name == "*"):
                new_cols.append(p)
        if proj:
            for p in proj.cols:
                if not (
                    hasattr(p, "table") and p.table == entity_table and p.name == "*"
                ):
                    new_cols.append(p)
        query.proj = Projection(new_cols)
    else:
        query.proj = proj


def _build_query(store, entity_table, qry, stmt):
    where = stmt.get("where")
    if where:
        where.set_table(entity_table)
        qry.append(where)
    attrs = stmt.get("attrs", "*")
    cols = attrs.split(",")
    _set_projection(store, entity_table, qry, cols)
    sort_by = stmt.get("attribute")
    if sort_by:
        direction = "ASC" if stmt["ascending"] else "DESC"
        qry.append(Order([(sort_by, direction)]))
    limit = stmt.get("limit")
    if limit:
        qry.append(Limit(limit))
    offset = stmt.get("offset")
    if offset:
        qry.append(Offset(offset))
    return qry


def _is_prefetch_allowed_in_config(pre_fetch_config, command_name, entity_type):
    if pre_fetch_config["switch_per_command"][
        command_name
    ] and entity_type not in lowered_str_list(pre_fetch_config["excluded_entities"]):
        return True
    else:
        return False



from kestrel.codegen.data import dump_data_to_file
from kestrel.codegen.summary import get_variable_entity_count
from kestrel.syntax.utils import get_all_input_var_names


class VarStruct:
    # variable data structure in symbol table

    def __init__(
        self,
        store,
        entity_table,
        link_tables,
        statement,
        entity_type,
        dep_vars,
        data_source,
        length=None,
        records_count=None,
    ):
        self.store = store

        # pointer (name of table) to internal data path (currently a view in SQLite)
        self.entity_table = entity_table

        # pointers (names of tables) to join tables for future many-to-many relations
        self.link_tables = link_tables

        # entity/SCO type of the variable
        self.type = entity_type

        # how many entities/SCOs in the variable
        if length is not None:
            self.length = length
        else:
            self.length = get_variable_entity_count(self)
        if records_count is not None:
            self.records_count = records_count
        elif self.entity_table:
            summary = self.store.summary(self.entity_table)
            self.records_count = summary["number_observed"]
        else:
            self.records_count = 0

        # TODO: cache of attributes for fast code completion request
        self.attributes = []

        # dependent variables
        self.dependent_variables = dep_vars

        # statement that generates the variable
        self.birth_statement = statement

        self.data_source = data_source

    def get_entities(self):
        return self.store.lookup(self.entity_table) if self.entity_table else []

    def dump_to_file(self, file_path):
        dump_data_to_file(self.store, self.entity_table, file_path)

    def __len__(self):
        return self.length

    def __repr__(self):
        return str(
            {
                "store name": self.store.dbname,
                "entity table name": self.entity_table,
                "link table names": self.link_tables,
                "entity type": self.type,
                "#(entities)": self.length,
                "#(records)": self.records_count,
                "attributes": self.attributes,
                "dependent variables": self.dependent_variables,
                "birth statement": self.birth_statement,
                "associated data source": self.data_source,
            }
        )

    def __iter__(self):
        """Useful for converting VarStruct to dict"""
        for attr in [
            "entity_table",
            "link_tables",
            "birth_statement",
            "type",
            "dependent_variables",
            "data_source",
            "length",
            "records_count",
        ]:
            yield attr, getattr(self, attr)


def _get_entity_type(stmt, symtable, dep_vars):
    if "type" in stmt:
        return stmt["type"]
    else:
        dep_var_name = dep_vars[0]
        dep_var = symtable[dep_var_name]
        return dep_var.type


def _get_data_source(stmt, dep_var_names, symtable):
    data_source = None
    if stmt["command"] == "get" and "datasource" in stmt:
        data_source = stmt["datasource"]
    else:
        dep_var_datasources = [symtable[v].data_source for v in dep_var_names]
        for ds in dep_var_datasources:
            if ds:
                data_source = ds
                break
    return data_source


def new_var(store, entity_table, link_tables, statement, symtable):
    dep_vars = get_all_input_var_names(statement)
    entity_type = _get_entity_type(statement, symtable, dep_vars)
    data_source = _get_data_source(statement, dep_vars, symtable)
    return VarStruct(
        store, entity_table, link_tables, statement, entity_type, dep_vars, data_source
    )



import os
import yaml
import pathlib
import pkgutil
import logging

from kestrel.utils import update_nested_dict

CONFIG_DIR_DEFAULT = pathlib.Path.home() / ".config" / "kestrel"
CONFIG_PATH_DEFAULT = CONFIG_DIR_DEFAULT / "kestrel.yaml"
CONFIG_PATH_ENV_VAR = "KESTREL_CONFIG"  # override CONFIG_PATH_DEFAULT if provided

_logger = logging.getLogger(__name__)


def load_default_config():
    _logger.debug(f"Loading default config file...")
    return yaml.safe_load(pkgutil.get_data(__name__, "config.yaml"))


def load_user_config(config_path_env_var, config_path_default):
    config_path = os.getenv(config_path_env_var, config_path_default)
    config_path = os.path.expanduser(config_path)
    config = {}
    if config_path:
        try:
            with open(config_path, "r") as fp:
                _logger.debug(f"User configuration file found: {config_path}")
                config = yaml.safe_load(fp)
        except FileNotFoundError:
            _logger.debug(f"User configuration file not exist.")
    return config


def load_config():
    config_default = load_default_config()
    config_user = load_user_config(CONFIG_PATH_ENV_VAR, CONFIG_PATH_DEFAULT)
    _logger.debug(f"User configuration loaded: {config_user}")
    _logger.debug(f"Updating default config with user config...")
    return update_nested_dict(config_default, config_user)



from firepit.exceptions import InvalidAttr
from firepit.query import (
    Query,
    Projection,
    Table,
    Filter,
    Predicate,
    Count,
    Unique,
    Join,
)
from collections import OrderedDict
from kestrel.codegen.relations import get_entity_id_attribute
from kestrel.exceptions import KestrelInternalError


def gen_variable_summary(var_name, var_struct):
    """Generate summary dictionary for a variable.

    Args:

        variable (kestrel.symboltable.variable.VarStruct): The target variable.

    Returns:

        dict: ``str`` to ``str`` dict on the variable summary.

    """
    summary = OrderedDict()
    footnote = "*Number of related records cached."

    summary["VARIABLE"] = var_name
    summary["TYPE"] = var_struct.type
    summary["#(ENTITIES)"] = len(var_struct)
    summary["#(RECORDS)"] = var_struct.records_count

    query_ids = _get_variable_query_ids(var_struct)

    is_from_direct_datasource = False
    var_birth_cmd = var_struct.birth_statement["command"]
    if var_birth_cmd == "find" or (
        var_birth_cmd == "get" and "datasource" in var_struct.birth_statement
    ):
        is_from_direct_datasource = True

    for entity_type in var_struct.store.types():

        if entity_type not in ("identity", "observed-data"):

            count = 0

            if query_ids and is_from_direct_datasource:
                query = Query(
                    [
                        Table(entity_type),
                        Join("__contains", "id", "=", "target_ref"),
                        Join("__queries", "source_ref", "=", "sco_id"),
                        Filter([Predicate("query_id", "IN", query_ids)]),
                        Unique(),
                        Count(),
                    ]
                )
                result = var_struct.store.run_query(query).fetchall()
                count = result[0]["count"]
                if entity_type == var_struct.type and count:
                    count = count - len(var_struct)
                    if count < 0:
                        raise KestrelInternalError(
                            f"impossible count regarding variable {var_name} and type {entity_type}"
                        )

            summary[f"{entity_type}*"] = count

    return summary, footnote


def _get_variable_query_ids(variable):
    query_ids = []
    if variable.entity_table:
        query = Query()
        query.append(Table("__queries"))
        query.append(Join(variable.entity_table, "sco_id", "=", "id"))
        query.append(Projection(["query_id"]))
        query.append(Unique())
        try:
            rows = variable.store.run_query(query).fetchall()
            query_ids = [r["query_id"] for r in rows]
        except InvalidAttr:
            pass
    return query_ids


def get_variable_entity_count(variable):
    entity_count = 0
    if variable.entity_table:
        entity_id_attr = get_entity_id_attribute(variable)
        if entity_id_attr not in variable.store.columns(variable.entity_table):
            return 0
        entity_count = variable.store.count(variable.entity_table)
    return entity_count



import dateutil
import logging
from typeguard import typechecked
from firepit.sqlstorage import SqlStorage
from kestrel.symboltable.symtable import SymbolTable
from kestrel.syntax.reference import Reference
from kestrel.exceptions import (
    InvalidAttribute,
    KestrelInternalError,
)

_logger = logging.getLogger(__name__)


@typechecked
def make_deref_func(store: SqlStorage, symtable: SymbolTable):
    def deref(reference: Reference):
        _logger.debug(f"deref {reference}")
        entity_table = symtable[reference.variable].entity_table

        try:
            store_return = store.lookup(entity_table, reference.attribute)
        except InvalidAttr as e:
            _logger.warning(f"cannot deref {reference}. Invalid attribute in firepit.")
            raise InvalidAttribute(e.message)

        if reference.attribute not in store_return[0]:
            _logger.warning(f"firepit does not return the correct column.")
            raise KestrelInternalError(
                f"firepit does not return {reference} correctly."
            )

        values = [row[reference.attribute] for row in store_return]

        # filter out None
        values = [v for v in values if v]

        # dedup
        values = tuple(set(values))

        _logger.debug(f"deref results: {str(values)}")

        return values

    return deref


@typechecked
def make_var_timerange_func(store: SqlStorage, symtable: SymbolTable):
    def get_timerange(reference: Reference):

        entity_table = symtable[reference.variable].entity_table

        summary = store.summary(entity_table)
        if summary["first_observed"] is None:
            start = None
        else:
            start = dateutil.parser.isoparse(summary["first_observed"])
        if summary["last_observed"] is None:
            end = None
        else:
            end = dateutil.parser.isoparse(summary["last_observed"])
        if start is None and end is None:
            tr = None
        else:
            tr = (start, end)

        return tr

    return get_timerange



import logging
import pathlib
import re

from kestrel.syntax.utils import get_all_input_var_names, timedelta_seconds
from kestrel.syntax.reference import deref_and_flatten_value_to_list
from kestrel.symboltable.symtable import SymbolTable
from firepit.sqlstorage import SqlStorage
from kestrel.datasource.manager import DataSourceManager

from kestrel.exceptions import (
    InvalidAttribute,
    VariableNotExist,
    UnsupportedRelation,
    KestrelInternalError,
    MissingDataSource,
)
from kestrel.codegen.relations import stix_2_0_ref_mapping, generic_relations
from kestrel.semantics.reference import make_deref_func, make_var_timerange_func

_logger = logging.getLogger(__name__)


def semantics_processing(
    stmt: dict,
    symtable: SymbolTable,
    store: SqlStorage,
    data_source_manager: DataSourceManager,
    config: dict,
):
    # semantics checking and completion

    _check_elements_not_empty(stmt)

    for input_var_name in get_all_input_var_names(stmt):
        _check_var_exists(input_var_name, symtable)

    if stmt["command"] == "get":
        _process_datasource_in_get(stmt, symtable, data_source_manager)

    if stmt["command"] == "load" or stmt["command"] == "save":
        stmt["path"] = pathlib.Path(stmt["path"]).expanduser().resolve()

    if stmt["command"] == "find":
        _check_semantics_on_find(stmt, symtable[stmt["input"]].type)

    if "attrs" in stmt:
        var_struct = symtable[stmt["input"]]
        stmt["attrs"] = _normalize_attrs(stmt, var_struct)

    deref_func = make_deref_func(store, symtable)
    get_timerange_func = make_var_timerange_func(store, symtable)

    if "where" in stmt:

        # 1. deref()
        stmt["where"].deref(deref_func, get_timerange_func)

        # 2. add_center_entity()
        if stmt["command"] in ("assign", "disp"):
            center_entity_type = symtable[stmt["input"]].type
        elif stmt["command"] in ("get", "find"):
            center_entity_type = stmt["type"]

        stmt["where"].add_center_entity(center_entity_type)

        # 3. to_stix() / to_firepit()
        if stmt["command"] in ("assign", "disp"):
            stmt["where"] = stmt["where"].to_firepit()
        elif stmt["command"] in ("get", "find"):
            time_adj = tuple(
                map(
                    timedelta_seconds,
                    (
                        config["stixquery"]["timerange_start_offset"],
                        config["stixquery"]["timerange_stop_offset"],
                    ),
                )
            )
            stmt["stixpattern"] = stmt["where"].to_stix(stmt["timerange"], time_adj)

    if "arguments" in stmt:
        stmt["arguments"] = {
            k: _arguments_deref_and_tostring(v, deref_func, get_timerange_func)
            for k, v in stmt["arguments"].items()
        }


def _check_elements_not_empty(stmt):
    for k, v in stmt.items():
        if isinstance(v, str) and not v:
            raise KestrelInternalError(f'incomplete parser; empty value for "{k}"')


def _check_var_exists(var_name, symtable):
    if var_name not in symtable:
        raise VariableNotExist(var_name)


def _normalize_attrs(stmt, v):
    props = []
    for attr in re.split(r",\s*", stmt["attrs"]):
        entity_type, _, prop = attr.rpartition(":")
        if entity_type and entity_type != v.type:
            raise InvalidAttribute(attr)
        props.append(prop)
    return ",".join(props)


def _process_datasource_in_get(stmt, symtable, data_source_manager):

    if stmt["command"] != "get":
        return

    # parser doesn't understand whether a data source is a Kestrel var
    # this function differente a Kestrel variable source from a data source
    if "datasource" in stmt:
        source = stmt["datasource"]
        if source in symtable:
            stmt["variablesource"] = source
            del stmt["datasource"]

    # complete default data source
    last_ds = data_source_manager.queried_data_sources[-1]
    if "variablesource" not in stmt and "datasource" not in stmt:
        if last_ds:
            stmt["datasource"] = last_ds
        else:
            raise MissingDataSource(stmt)


def _check_semantics_on_find(stmt, input_type):

    if stmt["command"] != "find":
        return

    # relation should be in lowercase after parsing by kestrel.syntax.parser.parse_kestrel()
    relation = stmt["relation"]
    return_type = stmt["type"]

    (entity_x, entity_y) = (
        (input_type, return_type) if stmt["reversed"] else (return_type, input_type)
    )

    if (
        entity_x,
        relation,
        entity_y,
    ) not in stix_2_0_ref_mapping and relation not in generic_relations:
        raise UnsupportedRelation(entity_x, relation, entity_y)


def _arguments_deref_and_tostring(v, deref_func, get_timerange_func):
    # not bother timerange for arguments deref
    w, _ = deref_and_flatten_value_to_list(v, deref_func, get_timerange_func)
    if len(w) == 1:
        w = str(w[0])
    else:
        w = ",".join(map(str, w))
    return w



from kestrel.absinterface import InterfaceManager
from kestrel.analytics import MODULE_PREFIX, AbstractAnalyticsInterface
from kestrel.exceptions import (
    AnalyticsInterfaceNotFound,
    InvalidAnalyticsInterfaceImplementation,
    ConflictingAnalyticsInterfaceScheme,
)


class AnalyticsManager(InterfaceManager):
    def __init__(self, config):
        super().__init__(
            config,
            "analytics",
            ["language", "default_analytics_schema"],
            MODULE_PREFIX,
            AbstractAnalyticsInterface,
            AnalyticsInterfaceNotFound,
            InvalidAnalyticsInterfaceImplementation,
            ConflictingAnalyticsInterfaceScheme,
        )

    def list_analytics_from_scheme(self, scheme):
        i, c = self._get_interface_with_config(scheme)
        return i.list_analytics(c)

    def execute(self, uri, argument_variables, session_id, parameters):
        scheme, uri = self._parse_and_complete_uri(uri)
        i, c = self._get_interface_with_config(scheme)
        rs = i.execute(uri, argument_variables, c, session_id, parameters)
        return rs



from firepit.query import (
    Column,
    Filter,
    Join,
    Predicate,
    Projection,
    Query,
    Table,
    Unique,
)

from kestrel.codegen.relations import stix_2_0_ref_mapping


def compile_specific_relation_to_query(
    return_type,
    relation,
    input_type,
    is_reversed,
    input_var_name,
    input_var_attrs,
    return_type_attrs,
):
    (entity_x, entity_y) = (
        (input_type, return_type) if is_reversed else (return_type, input_type)
    )

    stix_src_refs, stix_tgt_refs = stix_2_0_ref_mapping[(entity_x, relation, entity_y)]

    for ref_name in stix_src_refs:
        # e.g., # STIX: ("process", "created", "network-traffic"): (["opened_connection_refs"], [])
        #       # type(p) == process; is_reversed == True
        #       nt = FIND network-traffic CREATED BY p
        #       # type(nt) == network-traffic; is_reversed == False
        #       p = FIND process CREATED nt
        #
        # It is just aligned that is_reversed == whether input_var is
        # - EntityX in stix_2_0_ref_mapping
        # - the source_ref in the __reflist table of firepit v2.0
        var_is_source = is_reversed

        (var_attr, ret_attr) = (ref_name, "id") if var_is_source else ("id", ref_name)

        # if there are multiple options, use first one found in DB
        if ref_name.endswith("_refs"):
            query = _generate_reflist_query(
                input_var_name, var_is_source, ref_name, return_type
            )

        elif var_attr in input_var_attrs and ret_attr in return_type_attrs:
            query = _generate_ref_query(
                input_var_name, input_type, var_attr, return_type, ret_attr
            )

        else:
            continue

        return query

    for ref_name in stix_tgt_refs:
        # e.g., # STIX: ("autonomous-system", "owned", "ipv4-addr"): ([], ["belongs_to_refs"])
        #       # type(a) == autonomous-system; is_reversed == True
        #       ip = FIND ipv4-addr OWNED BY a
        #       # type(ip) == ipv4-addr; is_reversed == False
        #       a = FIND autonomous-system OWNED ip
        #
        # It is just aligned that (not is_reversed) == whether input_var is
        # - EntityY in stix_2_0_ref_mapping
        # - the source_ref in the __reflist table of firepit v2.0
        var_is_source = not is_reversed

        (var_attr, ret_attr) = (ref_name, "id") if var_is_source else ("id", ref_name)

        # if there are multiple options, use first one found in DB
        if ref_name.endswith("_refs"):
            query = _generate_reflist_query(
                input_var_name, var_is_source, ref_name, return_type
            )

        elif var_attr in input_var_attrs and ret_attr in return_type_attrs:
            query = _generate_ref_query(
                input_var_name, input_type, var_attr, return_type, ret_attr
            )

        else:
            continue

        return query

    return None


def compile_generic_relation_to_query(return_type, input_type, input_var_table):
    return SQLQuery(
        f"""
SELECT DISTINCT sco.*
 FROM __contains c
  JOIN "{return_type}" sco
   ON sco.id = c.target_ref
 WHERE source_ref IN (
  SELECT source_ref
   FROM __contains c
   WHERE target_ref LIKE '{return_type}--%'
  INTERSECT SELECT source_ref
   FROM __contains c
   WHERE target_ref IN (SELECT id FROM "{input_var_table}"))""",
        tuple(),
        input_var_table,
    )


# Utility class for overriding firepit behavior
# FIXME: A hack - need to figure out a btter way
class SQLQuery(Query):
    """A pre-written SQL query"""

    def __init__(self, query_text, query_values, var_name):
        super().__init__(var_name)
        self.text = query_text
        self.values = query_values

    def last_stage(self):
        return None

    def append(self, stage):
        raise NotImplementedError

    def render(self, placeholder, dialect=None):
        return self.text, self.values


def _generate_ref_query(input_var_name, input_type, var_attr, ret_type, ret_attr):
    # SELECT * FROM ret_type WHERE ret_type.ret_attr IN (SELECT var_attr FROM input_var_name)
    subq = Query(
        [
            Table(input_var_name),
            Projection([var_attr]),
        ]
    )
    return Query(
        [
            Table(ret_type),
            Filter([Predicate(ret_attr, "IN", subq)]),
        ]
    )


def _generate_reflist_query(input_var_name, var_is_source, ref_name, ret_type):
    var_ref_pos, ret_ref_pos = (
        ("source_ref", "target_ref") if var_is_source else ("target_ref", "source_ref")
    )
    return Query(
        [
            Table(input_var_name),
            Join("__reflist", "id", "=", var_ref_pos),
            Join(ret_type, ret_ref_pos, "=", "id"),
            Filter([Predicate("ref_name", "=", ref_name)]),
            Projection([Column("*", ret_type)]),  # All columns from ret_type
            Unique(),
        ]
    )



from datetime import datetime, timedelta
import dateutil
from pkgutil import get_data
import importlib
from lark import Lark, Token, Transformer
from lark.visitors import merge_transformers

from firepit.query import BinnedColumn
from kestrel.utils import unescape_quoted_string
from kestrel.syntax.ecgpattern import (
    ECGPComparison,
    ECGPJunction,
    ExtCenteredGraphPattern,
    Reference,
)

DEFAULT_VARIABLE = "_"
DEFAULT_SORT_ORDER = "DESC"


def parse_kestrel(
    stmts, default_variable=DEFAULT_VARIABLE, default_sort_order=DEFAULT_SORT_ORDER
):
    # the public parsing interface for Kestrel
    # return abstract syntax tree
    # check kestrel.lark for details
    grammar = get_data(__name__, "kestrel.lark").decode("utf-8")
    return Lark(
        grammar,
        parser="lalr",
        transformer=_KestrelT(default_variable, default_sort_order),
    ).parse(stmts)


def parse_ecgpattern(pattern_str) -> ExtCenteredGraphPattern:
    grammar = get_data(__name__, "ecgpattern.lark").decode("utf-8")
    paths = importlib.util.find_spec("kestrel.syntax").submodule_search_locations
    return Lark(
        grammar,
        parser="lalr",
        import_paths=paths,
        transformer=merge_transformers(_ECGPatternT(), kestrel=_KestrelT()),
    ).parse(pattern_str)


def parse_reference(value_str) -> Reference:
    grammar = get_data(__name__, "reference.lark").decode("utf-8")
    paths = importlib.util.find_spec("kestrel.syntax").submodule_search_locations
    parser = Lark(grammar, parser="lalr", import_paths=paths)

    try:
        ast = parser.parse(value_str)
    except:
        return None

    variable = ast.children[0].value
    attribute = ast.children[1].value

    return Reference(variable, attribute)


class _ECGPatternT(Transformer):
    def start(self, args):
        return ExtCenteredGraphPattern(args[0])


class _KestrelT(Transformer):
    def __init__(
        self, default_variable=DEFAULT_VARIABLE, default_sort_order=DEFAULT_SORT_ORDER
    ):
        self.default_variable = default_variable
        self.default_sort_order = default_sort_order
        super().__init__()

    def start(self, args):
        return args

    def statement(self, args):
        # Kestrel syntax: a statement can only has one command
        stmt = args.pop()
        return stmt

    def assignment(self, args):
        stmt = args[1] if len(args) == 2 else args[0]
        stmt["output"] = _extract_var(args, self.default_variable)
        return stmt

    def assign(self, args):
        packet = args[0]  # Already transformed in expression method below
        packet["command"] = "assign"
        return packet

    def merge(self, args):
        return {
            "command": "merge",
            "inputs": _extract_vars(args, self.default_variable),
        }

    def info(self, args):
        return {"command": "info", "input": _extract_var(args, self.default_variable)}

    def disp(self, args):
        packet = {"command": "disp"}
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        if "attrs" not in packet:
            packet["attrs"] = "*"
        return packet

    def get(self, args):
        packet = {
            "command": "get",
            "type": _extract_entity_type(args),
        }

        for item in args:
            if isinstance(item, dict):
                packet.update(item)

        if "timerange" not in packet:
            packet["timerange"] = None

        return packet

    def find(self, args):
        packet = {
            "command": "find",
            "type": _extract_entity_type(args),
            "relation": _assert_and_extract_single("RELATION", args).lower(),
            "reversed": _extract_if_reversed(args),
            "input": _extract_var(args, self.default_variable),
        }

        for item in args:
            if isinstance(item, dict):
                packet.update(item)

        if "timerange" not in packet:
            packet["timerange"] = None

        return packet

    def join(self, args):
        packet = {
            "command": "join",
            "input": _first(args),
            "input_2": _second(args),
        }
        if len(args) == 5:
            packet["attribute_1"] = _fourth(args)
            packet["attribute_2"] = _fifth(args)

        return packet

    def group(self, args):
        packet = {
            "command": "group",
            "attributes": args[2],
            "input": _extract_var(args, self.default_variable),
        }
        aggregations = args[3] if len(args) > 3 else None
        if aggregations:
            packet["aggregations"] = aggregations
        return packet

    def sort(self, args):
        return {
            "command": "sort",
            "attribute": _extract_attribute(args),
            "input": _extract_var(args, self.default_variable),
            "ascending": _extract_direction(args, self.default_sort_order),
        }

    def apply(self, args):
        packet = {"command": "apply", "arguments": {}}
        for arg in args:
            if isinstance(arg, dict):
                if "variables" in arg:
                    packet["inputs"] = arg["variables"]
                else:
                    packet.update(arg)
        return packet

    def load(self, args):
        packet = {
            "command": "load",
            "type": _extract_entity_type(args),
        }
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        return packet

    def save(self, args):
        packet = {
            "command": "save",
            "input": _extract_var(args, self.default_variable),
        }
        for arg in args:
            if isinstance(arg, dict):
                packet.update(arg)
        return packet

    def new(self, args):
        return {
            "command": "new",
            "type": _extract_entity_type(args),
            "data": _assert_and_extract_single("VAR_DATA", args),
        }

    def expression(self, args):
        packet = args[0]
        for arg in args:
            packet.update(arg)
        return packet

    def transform(self, args):
        return {
            "input": _extract_var(args, self.default_variable),
            "transform": _assert_and_extract_single("TRANSFORM", args),
        }

    def where_clause(self, args):
        pattern = ExtCenteredGraphPattern(args[0])
        return {
            "where": pattern,
        }

    def expression_or(self, args):
        return ECGPJunction("OR", args[0], args[1])

    def expression_and(self, args):
        return ECGPJunction("AND", args[0], args[1])

    def comparison_std(self, args):
        etype, attr = _extract_entity_and_attribute(args[0].value)
        # remove more than one spaces; capitalize op
        op = " ".join(_second(args).split()).upper()
        value = args[2]
        return ECGPComparison(attr, op, value, etype)

    def comparison_null(self, args):
        etype, attr = _extract_entity_and_attribute(args[0].value)
        op = _second(args)
        if "NOT" in op:
            op = "!="
        else:
            op = "="
        value = "NULL"
        return ECGPComparison(attr, op, value, etype)

    def value(self, args):
        return args[0]

    def literal_list(self, args):
        if len(args) == 1:
            return args[0]
        else:
            return args

    def literal(self, args):
        if args[0].type == "NUMBER":
            try:
                v = int(args[0].value)
            except:
                v = float(args[0].value)
        elif args[0].type == "ESCAPED_STRING":
            v = unescape_quoted_string(args[0].value)
        else:
            v = args[0].value
            ref = parse_reference(v)
            if ref:
                v = ref
        return v

    def attr_clause(self, args):
        paths = _assert_and_extract_single("ATTRIBUTES", args)
        return {
            "attrs": paths if paths else "*",
        }

    def sort_clause(self, args):
        return {
            "attribute": _extract_attribute(args),
            "ascending": _extract_direction(args, self.default_sort_order),
        }

    def limit_clause(self, args):
        return {
            "limit": int(_first(args)),
        }

    def offset_clause(self, args):
        return {
            "offset": int(_first(args)),
        }

    def timespan_relative(self, args):
        num = int(args[0])
        unit = args[1]
        if unit.type == "DAY":
            delta = timedelta(days=num)
        elif unit.type == "HOUR":
            delta = timedelta(hours=num)
        elif unit.type == "MINUTE":
            delta = timedelta(minutes=num)
        elif unit.type == "SECOND":
            delta = timedelta(seconds=num)
        stop = datetime.utcnow()
        start = stop - delta
        return {"timerange": (start, stop)}

    def timespan_absolute(self, args):
        start = dateutil.parser.isoparse(args[0])
        stop = dateutil.parser.isoparse(args[1])
        return {"timerange": (start, stop)}

    def timestamp(self, args):
        return _assert_and_extract_single("ISOTIMESTAMP", args)

    def entity_type(self, args):
        return _first(args)

    def variables(self, args):
        return {"variables": _extract_vars(args, self.default_variable)}

    def stdpath(self, args):
        v = _first(args)
        if args[0].type == "PATH_ESCAPED":
            v = unescape_quoted_string(v)
        return {"path": v}

    def datasource(self, args):
        v = _first(args)
        if args[0].type == "DATASRC_ESCAPED":
            v = unescape_quoted_string(v)
        return {"datasource": v}

    def analytics_uri(self, args):
        v = _first(args)
        if args[0].type == "ANALYTICS_ESCAPED":
            v = unescape_quoted_string(v)
        return {"analytics_uri": v}

    # automatically put one or more grp_expr into a list
    def grp_spec(self, args):
        return args

    def grp_expr(self, args):
        item = args[0]
        if isinstance(item, Token):
            # an ATTRIBUTE
            return str(item)
        else:
            # bin_func
            return item

    def bin_func(self, args):
        attr = _first(args)
        num = int(_second(args))
        if len(args) >= 3:
            unit = _third(args)
        else:
            unit = None
        alias = f"{attr}_bin"
        return BinnedColumn(attr, num, unit, alias=alias)

    def agg_list(self, args):
        return [arg for arg in args]

    def agg(self, args):
        func = _first(args).lower()
        attr = _second(args)
        alias = _third(args) if len(args) > 2 else f"{func}_{attr}"
        return {"func": func, "attr": attr, "alias": alias}

    def args(self, args):
        d = {}
        for di in args:
            d.update(di)
        return {"arguments": d}

    def arg_kv_pair(self, args):
        return {_first(args): args[1]}


def _first(args):
    return args[0].value


def _second(args):
    return args[1].value


def _third(args):
    return args[2].value


def _fourth(args):
    return args[3].value


def _fifth(args):
    return args[4].value


def _last(args):
    return args[-1].value


def _assert_and_extract_single(arg_type, args):
    items = [arg.value for arg in args if hasattr(arg, "type") and arg.type == arg_type]
    assert len(items) <= 1
    return items.pop() if items else None


def _extract_var(args, default_variable):
    # extract a single variable from the args
    # default variable if no variable is found
    v = _assert_and_extract_single("VARIABLE", args)
    return v if v else default_variable


def _extract_vars(args, default_variable):
    var_names = []
    for arg in args:
        if hasattr(arg, "type") and arg.type == "VARIABLE":
            var_names.append(arg.value)
    if not var_names:
        var_names = [default_variable]
    return var_names


def _extract_attribute(args):
    # extract a single attribute from the args
    return _assert_and_extract_single("ATTRIBUTE", args)


def _extract_entity_type(args):
    # extract a single entity type from the args
    return _assert_and_extract_single("ENTITY_TYPE", args)


def _extract_direction(args, default_sort_order):
    # extract sort direction from args
    # default direction if no variable is found
    # return: if descending
    ds = [
        x for x in args if hasattr(x, "type") and (x.type == "ASC" or x.type == "DESC")
    ]
    assert len(ds) <= 1
    d = ds.pop().type if ds else default_sort_order
    return True if d == "ASC" else False


def _extract_if_reversed(args):
    rs = [x for x in args if hasattr(x, "type") and x.type == "REVERSED"]
    return True if rs else False


def _extract_entity_and_attribute(s):
    if ":" in s:
        etype, _, attr = s.partition(":")
    else:
        etype, attr = None, s
    return etype, attr
